from .conftest import base_config
import multiprocessing
import openamundsen as oa


def run_model(model):
    model.configure_logger()
    model.run()


def test_multiprocessing():
    config = base_config()
    config.end_date = '2020-01-16'
    config.output_data.timeseries.format = 'memory'
    config.output_data.grids.format = 'memory'

    config.snow.model = 'multilayer'
    model1 = oa.OpenAmundsen(config)
    model1.initialize()

    config.snow.model = 'cryolayers'
    model2 = oa.OpenAmundsen(config)
    model2.initialize()

    spawn_context = multiprocessing.get_context('spawn')
    with spawn_context.Pool() as pool:
        pool.map(run_model, [model1, model2])



from loguru import logger
import numpy as np
from .clearsky import _clear_sky_shortwave_irradiance
from openamundsen import (
    constants,
    interpolation,
    meteo,
    modules,
)
import xarray as xr


def clear_sky_shortwave_irradiance(model):
    roi = model.grid.roi

    if model.sun_params['sun_over_horizon']:
        mean_surface_albedo = model.state.surface.albedo[roi].mean()
        if np.isnan(mean_surface_albedo):
            # E.g. in the first timestep no albedo has yet been calculated; assume a default value here
            mean_surface_albedo = model.config.soil.albedo

        logger.debug('Calculating shadows')

        if model.grid.extended_grid.available:
            shadows_dem = model.grid.extended_grid.dem
        else:
            shadows_dem = model.state.base.dem

        # Downsample DEM before calculating shadows if requested
        shadows_downsampling_factor = model.config.meteo.radiation.shadows_downsampling_factor
        if shadows_downsampling_factor > 1:
            orig_shadows_shape = shadows_dem.shape
            shadows_dem = shadows_dem[::shadows_downsampling_factor, ::shadows_downsampling_factor]

        shadows = modules.radiation.shadows(
            shadows_dem,
            model.grid.resolution * shadows_downsampling_factor,
            model.sun_params['sun_vector'],
            num_sweeps=model.config.meteo.radiation.num_shadow_sweeps,
        )

        # Upsample shadows array if necessary (by simple repetition of array elements)
        if shadows_downsampling_factor > 1:
            shadows = (
                shadows
                .repeat(shadows_downsampling_factor, axis=0)
                .repeat(shadows_downsampling_factor, axis=1)
            )
            if shadows.shape[0] > orig_shadows_shape[0]:
                shadows = shadows[:orig_shadows_shape[0], :]
            if shadows.shape[1] > orig_shadows_shape[1]:
                shadows = shadows[:, :orig_shadows_shape[1]]

        if model.grid.extended_grid.available:
            model.grid.extended_grid.shadows = shadows
            shadows = shadows[
                model.grid.extended_grid.row_slice,
                model.grid.extended_grid.col_slice,
            ]

        logger.debug('Calculating clear-sky shortwave irradiance')
        dir_irr, diff_irr = _clear_sky_shortwave_irradiance(
            model.sun_params['day_angle'],
            model.sun_params['sun_vector'],
            shadows,
            model.state.base.dem,
            model.state.base.svf,
            model.state.base.normal_vec,
            model.state.meteo.atmos_press,
            model.state.meteo.precipitable_water,
            mean_surface_albedo,
            roi=roi,
            ozone_layer_thickness=model.config.meteo.radiation.ozone_layer_thickness,
            atmospheric_visibility=model.config.meteo.radiation.atmospheric_visibility,
            single_scattering_albedo=model.config.meteo.radiation.single_scattering_albedo,
            clear_sky_albedo=model.config.meteo.radiation.clear_sky_albedo,
        )
    else:
        dir_irr = np.zeros(model.grid.shape)
        diff_irr = np.zeros(model.grid.shape)

    pot_irr = dir_irr + diff_irr
    model.state.meteo.sw_in_clearsky[roi] = pot_irr[roi]
    model.state.meteo.dir_in_clearsky[roi] = dir_irr[roi]
    model.state.meteo.diff_in_clearsky[roi] = diff_irr[roi]


def extended_grid_stations_clear_sky_shortwave_irradiance(model):
    logger.debug('Calculating clear-sky shortwave irradiance for extended-grid stations')
    roi = model.grid.roi
    ext_grid = model.grid.extended_grid

    ext_row_offset = ext_grid.row_offset
    ext_col_offset = ext_grid.col_offset
    station_rows_ext = model.meteo.row + ext_row_offset
    station_cols_ext = model.meteo.col + ext_col_offset
    # TODO this is the same in every time step, maybe add a "within_extended_grid_extent"
    # variable to model.meteo
    stations_within_ext_grid = (
        (station_rows_ext >= 0)
        & (station_cols_ext >= 0)
        & (station_rows_ext < ext_grid.rows)
        & (station_cols_ext < ext_grid.cols)
    )

    meteo_ds = (
        model.meteo
        .sel(station=stations_within_ext_grid & ~model.meteo.within_grid_extent)
        .sel(time=model.date)
    )
    # Consider only stations with temperature and humidity measurements (required for calculating
    # precipitable water)
    meteo_ds = meteo_ds.dropna(dim='station', how='any', subset=['temp', 'rel_hum'])
    clear_sky_rad = xr.DataArray(
        np.full(meteo_ds.sizes['station'], np.nan),
        coords=meteo_ds.coords,
        dims=meteo_ds.dims,
    )

    if meteo_ds.sizes['station'] > 0:
        extgrid_rows = (meteo_ds.row + ext_grid.row_offset).values
        extgrid_cols = (meteo_ds.col + ext_grid.col_offset).values

        # This is also calculated twice (here and in clear_sky_shortwave_irradiance() :/)
        mean_surface_albedo = model.state.surface.albedo[roi].mean()
        if np.isnan(mean_surface_albedo):
            mean_surface_albedo = model.config.soil.albedo

        extgrid_atmos_press = meteo.atmospheric_pressure(
            ext_grid.dem[extgrid_rows, extgrid_cols],
        )
        extgrid_precipitable_water = meteo.precipitable_water(
            meteo_ds.temp,
            meteo.vapor_pressure(meteo_ds.temp, meteo_ds.rel_hum),
        ).values
        # (-> this works only for stations with temperature and humidity measurements)

        extgrid_dir_irr, extgrid_diff_irr = _clear_sky_shortwave_irradiance(
            model.sun_params['day_angle'],
            model.sun_params['sun_vector'],
            ext_grid.shadows[extgrid_rows, extgrid_cols],
            ext_grid.dem[extgrid_rows, extgrid_cols],
            ext_grid.svf[extgrid_rows, extgrid_cols],
            ext_grid.normal_vec[:, extgrid_rows, extgrid_cols],
            extgrid_atmos_press,
            extgrid_precipitable_water,
            mean_surface_albedo,
            ozone_layer_thickness=model.config.meteo.radiation.ozone_layer_thickness,
            atmospheric_visibility=model.config.meteo.radiation.atmospheric_visibility,
            single_scattering_albedo=model.config.meteo.radiation.single_scattering_albedo,
            clear_sky_albedo=model.config.meteo.radiation.clear_sky_albedo,
        )
        clear_sky_rad[:] = extgrid_dir_irr + extgrid_diff_irr

    return clear_sky_rad


def shortwave_irradiance(model):
    logger.debug('Calculating actual shortwave irradiance')
    cloud_config = model.config.meteo.interpolation.cloudiness
    roi = model.grid.roi
    m = model.state.meteo
    ext_grid = model.grid.extended_grid
    method = cloud_config['method']

    meteo_ds = model.meteo.sel(time=model.date)

    if method == 'clear_sky_fraction':
        # Select stations within the grid extent and with shortwave radiation measurements for the
        # current time step
        ds_rad = (
            meteo_ds
            .sel(station=model.meteo.within_grid_extent)
            .dropna('station', subset=['sw_in'])
        )
        num_rad_stations = len(ds_rad.station)

    # Select stations with temperature and humidity measurements for the current time step
    ds_temp_hum = meteo_ds.dropna('station', how='any', subset=['temp', 'rel_hum'])
    num_temp_hum_stations = len(ds_temp_hum.station)

    if method == 'clear_sky_fraction':
        if model.sun_params['sun_over_horizon']:
            if ext_grid.available:
                extgrid_sw_in_clearsky = extended_grid_stations_clear_sky_shortwave_irradiance(
                    model
                )
                rad_ds_extgrid = (
                    meteo_ds
                    .sel(station=extgrid_sw_in_clearsky.station)
                    .dropna('station', subset=['sw_in'])
                )
                num_rad_stations += rad_ds_extgrid.sizes['station']
        else:
            method = cloud_config['clear_sky_fraction_night_method']

    if cloud_config['allow_fallback']:
        if method == 'clear_sky_fraction' and num_rad_stations == 0:
            logger.debug(
                'No radiation measurements available, trying humidity-based cloudiness calculation'
            )
            method = 'humidity'

        if method == 'humidity' and num_temp_hum_stations == 0:
            logger.debug(
                'No temperature/humidity measurements available, using cloudiness from previous '
                'time step'
            )
            method = 'constant'

    interpolate_cloud_factor = False

    if method == 'constant':  # use cloudiness from previous time step
        cloud_factor_roi_prev = m.cloud_factor[roi]

        # When there is no cloudiness from the previous time step (e.g. when the model run starts
        # during nighttime) set cloudiness to a constant value
        logger.debug(
            'No cloudiness from previous time step available, setting to constant value'
        )
        cloud_factor_roi_prev[np.isnan(cloud_factor_roi_prev)] = 0.75

        m.cloud_factor[roi] = cloud_factor_roi_prev
    elif method == 'humidity':
        lr_t = model.config.meteo.interpolation.temperature.lapse_rate[model.date.month - 1]
        lr_td = model.config.meteo.interpolation.humidity.lapse_rate[model.date.month - 1]
        # TODO use here also the same settings for regression/fixed gradients as for the
        # interpolation
        cloud_fracs = meteo.cloud_fraction_from_humidity(
            ds_temp_hum.temp,
            ds_temp_hum.rel_hum,
            ds_temp_hum.alt,
            lr_t,
            lr_td,
            pressure_level=cloud_config.pressure_level * 100,  # hPa -> Pa
            saturation_cloud_fraction=cloud_config.saturation_cloud_fraction,
            e_folding_humidity=cloud_config.e_folding_humidity,
        )
        cloud_factor_xs = ds_temp_hum.x
        cloud_factor_ys = ds_temp_hum.y
        cloud_factors = meteo.cloud_factor_from_cloud_fraction(cloud_fracs)
        interpolate_cloud_factor = True
    elif method == 'clear_sky_fraction':
        cloud_factor_xs = ds_rad.x
        cloud_factor_ys = ds_rad.y
        cloud_factors = (
            ds_rad.sw_in.values / m.sw_in_clearsky[ds_rad.row, ds_rad.col]
        )
        interpolate_cloud_factor = True

        if ext_grid.available:
            extgrid_cloud_factors = rad_ds_extgrid.sw_in / extgrid_sw_in_clearsky
            cloud_factor_xs = np.append(cloud_factor_xs, rad_ds_extgrid.x)
            cloud_factor_ys = np.append(cloud_factor_ys, rad_ds_extgrid.y)
            cloud_factors = np.append(cloud_factors, extgrid_cloud_factors)
    elif method == 'prescribed':
        cloudiness_ds = meteo_ds.dropna('station', subset=['cloud_fraction'])
        cloud_factor_xs = cloudiness_ds.x
        cloud_factor_ys = cloudiness_ds.y
        cloud_fractions = cloudiness_ds.cloud_fraction / 100  # convert from % to 0-1
        cloud_factors = meteo.cloud_factor_from_cloud_fraction(cloud_fractions)
        interpolate_cloud_factor = True

    if interpolate_cloud_factor:
        cloud_factor_interpol = interpolation.idw(
            cloud_factor_xs,
            cloud_factor_ys,
            cloud_factors,
            model.grid.roi_points[:, 0],
            model.grid.roi_points[:, 1],
        )
        m.cloud_factor[roi] = cloud_factor_interpol.clip(0, 1)

    m.cloud_fraction[roi] = meteo.cloud_fraction_from_cloud_factor(m.cloud_factor[roi])
    m.sw_in[roi] = m.sw_in_clearsky[roi] * m.cloud_factor[roi]


def longwave_irradiance(model):
    logger.debug('Calculating longwave irradiance')
    roi = model.grid.roi
    m = model.state.meteo
    clear_sky_emissivity = meteo.clear_sky_emissivity(m.precipitable_water[roi])

    # TODO these should be parameters
    cloud_emissivity = 0.976  # emissivity of totally overcast skies (Greuell et al., 1997)
    rock_emission_factor = 0.01  # (K W-1 m2) temperature of emitting rocks during daytime is assumed to be higher than the air temperature by this factor multiplied by the incoming shortwave radiation (Greuell et al., 1997)

    # Incoming longwave radiation from the clear sky
    lw_in_clearsky = (
        clear_sky_emissivity
        * constants.STEFAN_BOLTZMANN
        * m.temp[roi]**4
        * model.state.base.svf[roi]
        * (1 - m.cloud_fraction[roi]**2)
    )

    # Incoming longwave radiation from clouds
    lw_in_clouds = (
        cloud_emissivity
        * constants.STEFAN_BOLTZMANN
        * m.temp[roi]**4
        * model.state.base.svf[roi]
        * m.cloud_fraction[roi]**2
    )

    # Incoming longwave radiation from surrounding slopes
    snowfree_count = (model.state.snow.swe[roi] == 0).sum()
    rock_fraction = snowfree_count / model.grid.roi.sum()
    lw_in_slopes = (
        constants.STEFAN_BOLTZMANN
        * (1 - model.state.base.svf[roi]) * (
            rock_fraction * (m.temp[roi] + rock_emission_factor * m.sw_in[roi])**4
            + (1 - rock_fraction) * model.state.surface.temp[roi]**4
        )
    )

    # Total incoming longwave radiation
    m.lw_in[roi] = lw_in_clearsky + lw_in_clouds + lw_in_slopes



from openamundsen import constants as c
import numpy as np


# Magnus formula coefficients a (Pa), b, and c (°C)
VAPOR_PRESSURE_COEFFS_ICE = (610.71, 22.44, 272.44)
VAPOR_PRESSURE_COEFFS_WATER = (610.78, 17.08, 234.18)


def atmospheric_pressure(elev):
    """
    Calculate atmospheric pressure.

    Parameters
    ----------
    elev : numeric
        Height above sea level (m).

    Returns
    -------
    pressure : numeric
        Atmospheric pressure (Pa).
    """
    return c.STANDARD_ATMOSPHERE * (
        1 - c.GRAVITATIONAL_ACCELERATION * np.asarray(elev)
        / (c.SPEC_HEAT_CAP_DRY_AIR * c.STANDARD_SEA_LEVEL_TEMPERATURE)
    )**(c.SPEC_HEAT_CAP_DRY_AIR * c.MOLAR_MASS_DRY_AIR / c.UNIVERSAL_GAS_CONSTANT)


def pressure_to_altitude(pressure):
    """
    Calculate altitude from atmospheric pressure.

    This is the inverse function for atmospheric_pressure().

    Parameters
    ----------
    pressure : numeric
        Atmospheric pressure (Pa).

    Returns
    -------
    elev : numeric
        Height above sea level (m).
    """
    exp = c.SPEC_HEAT_CAP_DRY_AIR * c.MOLAR_MASS_DRY_AIR / c.UNIVERSAL_GAS_CONSTANT
    return (
        (1 - (np.asarray(pressure) / c.STANDARD_ATMOSPHERE)**(1 / exp))
        * c.SPEC_HEAT_CAP_DRY_AIR * c.STANDARD_SEA_LEVEL_TEMPERATURE
        / c.GRAVITATIONAL_ACCELERATION
    )


def dry_air_density(temp, pressure):
    """
    Calculate the density of dry air.

    Parameters
    ----------
    temp : numeric
        Air temperature (K).

    pressure : numeric
        Atmospheric pressure (Pa).

    Returns
    -------
    density : numeric
        Dry air density (kg m-3).
    """
    return pressure * (c.GAS_CONSTANT_DRY_AIR * temp)


def latent_heat_of_vaporization(temp):
    """
    Calculate the latent heat of vaporization.

    Parameters
    ----------
    temp : numeric
        Air temperature (K).

    Returns
    -------
    lat_heat_vap : numeric
        Latent heat of vaporization (J kg-1).
    """
    return c.LATENT_HEAT_OF_VAPORIZATION - 2.361e3 * (np.asarray(temp) - c.T0)


def saturation_vapor_pressure(temp, over=None):
    """
    Calculate saturation vapor pressure for a given temperature.

    Parameters
    ----------
    temp : numeric
        Air temperature (K).

    over : str, default None
        Can be 'water' (to calculate vapor pressure over water), 'ice' (to
        calculate vapor pressure over ice), or None (to decide depending on
        temperature).

    Returns
    -------
    sat_vapor_pressure : numeric
        Saturation vapor pressure (Pa).
    """
    temp_c = np.asarray(temp) - c.T0

    if over is None:
        pos = temp_c >= 0
        ca = np.where(pos, VAPOR_PRESSURE_COEFFS_WATER[0], VAPOR_PRESSURE_COEFFS_ICE[0])
        cb = np.where(pos, VAPOR_PRESSURE_COEFFS_WATER[1], VAPOR_PRESSURE_COEFFS_ICE[1])
        cc = np.where(pos, VAPOR_PRESSURE_COEFFS_WATER[2], VAPOR_PRESSURE_COEFFS_ICE[2])
    elif over == 'water':
        ca = VAPOR_PRESSURE_COEFFS_WATER[0]
        cb = VAPOR_PRESSURE_COEFFS_WATER[1]
        cc = VAPOR_PRESSURE_COEFFS_WATER[2]
    elif over == 'ice':
        ca = VAPOR_PRESSURE_COEFFS_ICE[0]
        cb = VAPOR_PRESSURE_COEFFS_ICE[1]
        cc = VAPOR_PRESSURE_COEFFS_ICE[2]
    else:
        raise NotImplementedError

    return ca * np.exp(cb * temp_c / (cc + temp_c))


def vapor_pressure(temp, rel_hum, over=None):
    """
    Calculate vapor pressure.

    Parameters
    ----------
    temp : numeric
        Air temperature (K).

    rel_hum : numeric
        Relative humidity (%).

    over : str, default None
        Can be 'water' (to calculate vapor pressure over water), 'ice' (to
        calculate vapor pressure over ice), or None (to decide depending on
        temperature).

    Returns
    -------
    vapor_pressure : numeric
        Vapor pressure (Pa).
    """
    return saturation_vapor_pressure(temp, over=over) * np.asarray(rel_hum) / 100


def specific_humidity(atmospheric_pressure, vapor_pressure):
    """
    Calculate specific humidity.

    Parameters
    ----------
    atmospheric_pressure : numeric
        Atmospheric pressure (Pa).

    vapor_pressure : numeric
        Vapor pressure (Pa).

    Returns
    -------
    spec_hum : numeric
        Specific humidity (kg kg-1).
    """
    atmospheric_pressure = np.asarray(atmospheric_pressure)
    vapor_pressure = np.asarray(vapor_pressure)
    return 0.622 * vapor_pressure / (atmospheric_pressure - 0.378 * vapor_pressure)


def absolute_humidity(temp, vapor_pressure):
    """
    Calculate absolute humidity.

    Parameters
    ----------
    temp : numeric
        Air temperature (K).

    vapor_pressure : numeric
        Vapor pressure (Pa).

    Returns
    -------
    abs_hum : numeric
        Absolute humidity (kg m-3).
    """
    return np.asarray(vapor_pressure) / (c.SPEC_GAS_CONSTANT_WATER_VAPOR * np.asarray(temp))


def relative_humidity(temp, abs_hum):
    """
    Calculate relative humidity from absolute humidity.

    Parameters
    ----------
    temp : numeric
        Air temperature (K).

    abs_hum : numeric
        Absolute humidity (kg m-3).

    Returns
    -------
    rel_hum : numeric
        Relative humidity (%).
    """
    temp = np.asarray(temp)
    abs_hum = np.asarray(abs_hum)
    vap_press = c.SPEC_GAS_CONSTANT_WATER_VAPOR * temp * abs_hum
    sat_vap_press = saturation_vapor_pressure(temp)
    return 100 * vap_press / sat_vap_press


def specific_heat_capacity_moist_air(spec_hum):
    """
    Calculate the specific heat capacity of moist air.

    Parameters
    ----------
    spec_hum : numeric
        Specific humidity (kg kg-1).

    Returns
    -------
    spec_heat_cap_moist_air : numeric
        Specific heat capacity of moist air (J kg-1 K-1).
    """
    return c.SPEC_HEAT_CAP_DRY_AIR * (1 + 0.84 * np.asarray(spec_hum))


def psychrometric_constant(atmos_press, spec_heat_cap, lat_heat_vap):
    """
    Calculate the psychrometric constant.

    Parameters
    ----------
    atmos_press : numeric
        Atmospheric pressure (Pa).

    spec_heat_cap : numeric
        Specific heat capacity (J kg-1 K-1).

    lat_heat_vap : numeric
        Latent heat of vaporization (J kg-1).

    Returns
    -------
    psychrometric_constant : numeric
        Psychrometric constant (Pa K-1).
    """
    return np.asarray(spec_heat_cap) * np.asarray(atmos_press) / (0.622 * np.asarray(lat_heat_vap))


def _water_vapor_pressure_difference(temp, wet_bulb_temp, vap_press, psych_const):
    """
    Evaluate the psychrometric formula
        e_l - (e_w - gamma * (T_a - T_w)).

    Parameters
    ----------
    temp : numeric
        Air temperature (K).

    wet_bulb_temp : numeric
        Wet-bulb temperature (K).

    vap_press : numeric
        Vapor pressure (Pa).

    psych_const : numeric
        Psychrometric constant (Pa K-1).

    Returns
    -------
    wat_vap_press_diff : numeric
        Water vapor pressure difference (Pa).
    """
    sat_vap_press_wet_bulb = saturation_vapor_pressure(wet_bulb_temp)
    return vap_press - (sat_vap_press_wet_bulb - psych_const * (temp - wet_bulb_temp))


def wet_bulb_temperature(temp, rel_hum, vap_press, psych_const):
    """
    Calculate wet-bulb temperature for given ambient conditions.
    Wet-bulb temperature is calculated iteratively using a secant method.

    Parameters
    ----------
    temp : numeric
        Air temperature (K).

    rel_hum : numeric
        Relative humidity (%).

    vap_press : numeric
        Vapor pressure (Pa).

    psych_const : numeric
        Psychrometric constant (Pa K-1).

    Returns
    -------
    wet_bulb_temp : numeric
        Wet-bulb temperature (K).
    """
    temp = np.asarray(temp)
    rel_hum = np.asarray(rel_hum)
    vap_press = np.asarray(vap_press)
    psych_const = np.asarray(psych_const)

    tol = 1e-2  # stopping criterion (iteration continues until abs(x0 - x1) < tol)

    x0 = temp - 10  # first start value (x_(k-1))
    x1 = temp.copy()  # second start value (x_k)
    y0 = _water_vapor_pressure_difference(temp, x0, vap_press, psych_const)
    y1 = _water_vapor_pressure_difference(temp, x1, vap_press, psych_const)

    while True:
        d = (x1 - x0) / (y1 - y0) * y1  # secant method
        iter_pos = np.abs(d) > tol

        if iter_pos.sum() == 0:
            break

        x0[iter_pos] = x1[iter_pos]
        y0[iter_pos] = y1[iter_pos]
        x1[iter_pos] -= d[iter_pos]
        y1[iter_pos] = _water_vapor_pressure_difference(
            temp[iter_pos],
            x1[iter_pos],
            vap_press[iter_pos],
            psych_const[iter_pos],
        )

    return x1


def dew_point_temperature(temp, rel_hum):
    """
    Calculate dew point temperature.

    Parameters
    ----------
    temp : numeric
        Air temperature (K).

    rel_hum : numeric
        Relative humidity (%).

    Returns
    -------
    dew_point_temp : numeric
        Dew point temperature (K).
    """
    ca, cb, cc = VAPOR_PRESSURE_COEFFS_WATER
    vap_press_water = vapor_pressure(temp, rel_hum, 'water')
    td_c = cc * np.log(vap_press_water / ca) / (cb - np.log(vap_press_water / ca))
    return td_c + c.T0


def precipitable_water(temp, vap_press):
    """
    Calculate precipitable water after Prata (1996).

    Parameters
    ----------
    temp : numeric
        Air temperature (K).

    vap_press : numeric
        Vapor pressure (Pa).

    Returns
    -------
    precipitable_water : numeric
        Precipitable water (kg m-2).

    References
    ----------
    .. [1] Prata, A. J. (1996). A new long-wave formula for estimating downward
       clear-sky radiation at the surface. Quarterly Journal of the Royal
       Meteorological Society, 122(533), 1127–1151. doi:10.1002/qj.49712253306
    """
    vap_press_hpa = vap_press / 100
    u = 46.5 * vap_press_hpa / temp  # in g cm-2
    return u * (1e-3 * 1e4)  # in kg m-2


def cloud_fraction_from_humidity(
    temp,
    rel_hum,
    elev,
    temp_lapse_rate,
    dew_point_temp_lapse_rate,
    pressure_level=70000,
    saturation_cloud_fraction=0.832,
    e_folding_humidity=58.4,
):
    """
    Calculate cloud fraction from relative humidity at the 700 hPa level
    following Walcek (1994) and Liston and Elder (2006).

    Parameters
    ----------
    temp : numeric
        Air temperature (K).

    rel_hum : numeric
        Relative humidity (%).

    elev : numeric
        Elevation (m).

    temp_lapse_rate : numeric
        Temperature lapse rate (K m-1).

    dew_point_temp_lapse_rate : numeric
        Dew point temperature lapse rate (K m-1).

    pressure_level : float, default 70000
        Pressure level (Pa) on which the calculation should be performed.
        Default value of 700 hPa is from [2].

    saturation_cloud_fraction : float, default 0.832
        Cloud fraction at 100% relative humidity. Default value from [2].

    e_folding_humidity : float, default 58.4
        "e-folding relative humidity", i.e., the relative humidity depression
        below 100% where cloud fraction decreases to 37% (exp(-1)) of its value
        at 100% humidity. Default value from [2].

    Returns
    -------
    cloud_fraction : numeric
        Cloud fraction (0-1).

    References
    ----------
    .. [1] Walcek, C. J. (1994). Cloud Cover and Its Relationship to Relative
       Humidity during a Springtime Midlatitude Cyclone. Monthly Weather Review,
       122(6), 1021–1035.
       https://doi.org/10.1175/1520-0493(1994)122<1021:CCAIRT>2.0.CO;2

    .. [2] Liston, G. E., & Elder, K. (2006). A Meteorological Distribution
       System for High-Resolution Terrestrial Modeling (MicroMet). Journal of
       Hydrometeorology, 7(2), 217–234. https://doi.org/10.1175/JHM486.1
    """
    temp = np.asarray(temp)
    rel_hum = np.asarray(rel_hum)
    elev = np.asarray(elev)

    pressure_level_elev = pressure_to_altitude(pressure_level)
    elev_diff = elev - pressure_level_elev

    td = dew_point_temperature(temp, rel_hum)
    temp700 = temp - (temp_lapse_rate * elev_diff)  # T in 700 hPa
    td700 = td - (dew_point_temp_lapse_rate * elev_diff)  # Td in 700 hPa
    sat_vap_press700 = saturation_vapor_pressure(temp700)
    vap_press700 = saturation_vapor_pressure(td700)
    rel_hum700 = 100 * vap_press700 / sat_vap_press700  # RH in 700 hPa

    cloud_frac = (  # eq. (1) in [1], eq. (20) in [2]
        saturation_cloud_fraction
        * np.exp((rel_hum700 - 100) / (100 - e_folding_humidity))
    )
    return cloud_frac.clip(0, 1)


def cloud_factor_from_cloud_fraction(cloud_fraction):
    """
    Calculate the cloud factor (i.e., the ratio of the actual global radiation
    and the clear-sky global radiation) from the cloud fraction after Greuell
    et al. (1997).

    Parameters
    ----------
    cloud_fraction : numeric
        Cloud fraction (0-1).

    Returns
    -------
    cloud_factor : numeric
        Cloud factor (0-1).

    References
    ----------
    .. [1] Greuell, W., Knap, W. H., & Smeets, P. C. (1997). Elevational
       changes in meteorological variables along a midlatitude glacier during
       summer. Journal of Geophysical Research, 102(D22), 25941–25954.
       https://doi.org/10.1029/97JD02083
    """
    cloud_fraction = np.asarray(cloud_fraction)
    cloud_factor = 1 - 0.233 * cloud_fraction - 0.415 * cloud_fraction**2
    return cloud_factor.clip(0, 1)


def cloud_fraction_from_cloud_factor(cloud_factor):
    """
    Calculate cloud fraction from cloud factor after Greuell et al. (1997),
    inverted fit function.

    Parameters
    ----------
    cloud_factor : numeric
        Cloud factor (0-1).

    Returns
    -------
    cloud_fraction : numeric
        Cloud fraction (0-1).

    References
    ----------
    .. [1] Greuell, W., Knap, W. H., & Smeets, P. C. (1997). Elevational
       changes in meteorological variables along a midlatitude glacier during
       summer. Journal of Geophysical Research, 102(D22), 25941–25954.
       https://doi.org/10.1029/97JD02083
    """
    a = -0.415
    b = -0.233
    c = 1.
    cloud_fraction = (-b - np.sqrt(b**2 - 4 * a * (c - np.asarray(cloud_factor)))) / (2 * a)
    return cloud_fraction.clip(0, 1)


def clear_sky_emissivity(prec_wat):
    """
    Calculate emissivity of the atmosphere under clear-sky conditions after
    Prata (1996).

    Parameters
    ----------
    prec_wat : numeric
        Precipitable water (kg m-2).

    Returns
    -------
    emissivity : numeric
        Emissivity (0-1).

    References
    ----------
    .. [1] Prata, A. J. (1996). A new long-wave formula for estimating downward
       clear-sky radiation at the surface. Quarterly Journal of the Royal
       Meteorological Society, 122(533), 1127–1151. doi:10.1002/qj.49712253306
    """
    prec_wat_cm = np.asarray(prec_wat) / 10  # kg m-2 (= mm) to g cm-2 (= cm)
    return 1 - (1 + prec_wat_cm) * np.exp(-np.sqrt(1.2 + 3 * prec_wat_cm))


def precipitation_phase(temp, threshold_temp=c.T0, temp_range=0., method='linear'):
    """
    Calculate precipitation phase.

    Parameters
    ----------
    temp : numeric
        Temperature (in K) on which to perform the partitioning (e.g., air
        temperature or wet-bulb temperature).

    threshold_temp : float, default 273.15
        Threshold temperature (in K) at which 50% of precipitation falls as
        snow.

    temp_range : float, default 0
        Temperature range in which mixed precipitation can occur. If 0,
        precipitation can only be either rain or snow.

    method : str, default 'linear'
        Method for interpolating precipitation phase within the
        (threshold_temp - temp_range/2, threshold_temp + temp_range/2) range.

    Returns
    -------
    snowfall_fraction : numeric
        Fraction of precipitation falling as snow (0-1).
    """

    if method != 'linear':
        raise NotImplementedError

    temp = np.asarray(temp)

    if temp_range < 0:
        raise ValueError('temp_range must be positive')
    elif temp_range == 0:
        snowfall_frac = (temp < threshold_temp) * 1.
    else:
        t1 = threshold_temp - temp_range / 2.
        t2 = threshold_temp + temp_range / 2.
        snowfall_frac = (1 - (temp - t1) / (t2 - t1)).clip(0, 1)

    return snowfall_frac


def log_wind_profile(ref_wind_speed, ref_height, height, roughness_length):
    """
    Calculate wind speed at a different height based on a logarithmic wind
    profile.

    Parameters
    ----------
    ref_wind_speed : numeric
        Reference wind speed (m s-1).

    ref_height : float
        Height of the reference wind speed (m).

    height : float
        New height.

    roughness_length : float
        Surface roughness length (m).

    Returns
    -------
    wind_speed : numeric
        Wind speed at the new height (m s-1).
    """
    return ref_wind_speed * (
        np.log(height / roughness_length)
        / np.log(ref_height / roughness_length)
    )


def wind_to_uv(ws, wd):
    """
    Convert wind speed and direction to u and v components.

    Parameters
    ----------
    ws : numeric
        Wind speed (m s-1).

    wd : numeric
        Wind direction (meteorological degrees, i.e., 0 = North, 90 = East).

    Returns
    -------
    u : numeric
        Zonal component (m s-1).

    v : numeric
        Meridional component (m s-1).
    """
    u = -ws * np.sin(np.deg2rad(wd))
    v = -ws * np.cos(np.deg2rad(wd))
    return u, v


def wind_from_uv(u, v):
    """
    Convert u and v components to wind speed and direction.

    Parameters
    ----------
    u : numeric
        Zonal component (m s-1).

    v : numeric
        Meridional component (m s-1).

    Returns
    -------
    ws : numeric
        Wind speed (m s-1).

    wd : numeric
        Wind direction (meteorological degrees, i.e., 0 = North, 90 = East).
    """
    ws = np.sqrt(u**2 + v**2)
    wd = (270 - np.rad2deg(np.arctan2(v, u))) % 360
    return ws, wd



from numba import njit, prange
import numpy as np
from openamundsen import (
    constants as c,
)


def albedo(model, pos=None):
    """
    Update snow albedo.
    Albedo aging is calculated using an exponential decay function with
    different timescales for cold and melting snow [1-3].
    For fresh snow, albedo is either reset to the maximum value when snowfall
    exceeds a preset threshold [2], or continuously increased whenever snow is
    falling with a rate depending on snowfall rate [3-4].

    References
    ----------
    .. [1] Snow Hydrology: Summary Report of the Snow Investigations. Published
       by the North Pacific Division, Corps of Engineers, U.S. Army, Portland,
       Oregon, 1956.  437 pages, 70 pages of plates, maps and figs., 27 cm.
       https://doi.org/10.3189/S0022143000024503

    .. [2] Hanzer, F., Helfricht, K., Marke, T., & Strasser, U. (2016).
       Multilevel spatiotemporal validation of snow/ice mass balance and runoff
       modeling in glacierized catchments. The Cryosphere, 10(4), 1859–1881.
       https://doi.org/10.5194/tc-10-1859-2016

    .. [3] Essery, R., Morin, S., Lejeune, Y., & B Ménard, C. (2013). A
       comparison of 1701 snow models using observations from an alpine site.
       Advances in Water Resources, 55, 131–148.
       https://doi.org/10.1016/j.advwatres.2012.07.013

    .. [4] Dutra, E., Balsamo, G., Viterbo, P., Miranda, P., Beljaars, A.,
       Schär, C., & Elder, K. (2009). New snow scheme in HTESSEL: description and
       offline validation. ECMWF. https://doi.org/10.21957/98x9mrv1y
    """
    albedo_config = model.config.snow.albedo
    s = model.state

    if pos is None:
        pos = model.grid.roi

    if albedo_config.decay_timescale_determination_temperature == 'air':
        temp = s.meteo.temp[pos]
    elif albedo_config.decay_timescale_determination_temperature == 'surface':
        temp = s.surface.temp[pos]

    # Albedo decay
    decay_timescale = np.where(
        temp >= c.T0,
        c.SECONDS_PER_HOUR * albedo_config.melting_snow_decay_timescale,
        c.SECONDS_PER_HOUR * albedo_config.cold_snow_decay_timescale,
    )
    new_albedo = (
        albedo_config.min
        + (s.snow.albedo[pos] - albedo_config.min)
        * np.exp(-1 / decay_timescale * model.timestep)
    )

    # Albedo refresh
    snowfall_rate = s.meteo.snowfall[pos] / model.timestep
    refresh_snowfall_rate = albedo_config.refresh_snowfall / c.SECONDS_PER_HOUR  # kg m-2 s-1
    if albedo_config.refresh_method == 'binary':
        new_albedo[snowfall_rate >= refresh_snowfall_rate] = albedo_config.max
    elif albedo_config.refresh_method == 'continuous':
        new_albedo += (
            (albedo_config.max - new_albedo)
            * np.minimum(1, snowfall_rate / refresh_snowfall_rate)  # minimum() to clip values at albedo_config.max
        )

    s.snow.albedo[pos] = new_albedo


def fresh_snow_density(temp):
    """
    Calculate fresh snow density based on the parameterization by [1] (eq. (4.22)).

    Parameters
    ----------
    temp : numeric
        Wet-bulb temperature (K) in the original formulation by [1]. Can be
        substituted by air temperature if not available.

    Returns
    -------
    rho : numeric
        Fresh snow density (kg m-3).

    References
    ----------
    .. [1] Anderson, E. A. (1976). A point energy and mass balance model of a
       snow cover (NOAA Technical Report NWS 19, pp. 1–172). NOAA.
       https://repository.library.noaa.gov/view/noaa/6392
    """
    min_temp = c.T0 - 15.
    temp = np.array(temp).clip(min_temp)  # the parameterization is only valid for temperatures >= -15 °C
    rho = 50 + 1.7 * (temp - min_temp)**1.5
    rho[np.isnan(rho)] = 100.  # if temperature is not available, set density to 100 kg m-3
    return rho


def compaction(model):
    snow = model.state.snow

    if model.config.snow.compaction.method == 'anderson':
        _compaction_anderson(
            model.grid.roi_idxs,
            model.timestep,
            snow.num_layers,
            snow.thickness,
            snow.ice_content,
            snow.liquid_water_content,
            snow.density,
            model.state.meteo.temp,
        )
    elif model.config.snow.compaction.method == 'empirical':
        timescale = c.SECONDS_PER_HOUR * model.config.snow.compaction.timescale  # snow compaction timescale (s)
        _compaction_empirical(
            model.grid.roi_idxs,
            model.timestep,
            timescale,
            model.config.snow.compaction.max_cold_density,
            model.config.snow.compaction.max_melting_density,
            snow.num_layers,
            snow.temp,
            snow.thickness,
            snow.ice_content,
            snow.liquid_water_content,
            snow.density,
        )
    else:
        raise NotImplementedError


@njit(cache=True, parallel=True)
def _compaction_anderson(
    roi_idxs,
    timestep,
    num_layers,
    thickness,
    ice_content,
    liquid_water_content,
    density,
    air_temp,
):
    """
    Calculate snow compaction following [1].
    Parameter names (c1...c7) are changed compared to the original
    formulation, follwing [2].

    Parameters
    ----------
    roi_idxs : ndarray(int, ndim=2)
        (N, 2)-array specifying the (row, col) indices within the data arrays
        that should be considered.

    timestep : float
        Model timestep (s).

    num_layers : ndarray(float, ndim=2)
        Number of snow layers.

    thickness : ndarray(float, ndim=3)
        Snow thickness (m).

    ice_content : ndarray(float, ndim=3)
        Ice content of snow (kg m-2).

    liquid_water_content : ndarray(float, ndim=3)
        Liquid water content of snow (kg m-2).

    density : ndarray(float, ndim=3)
        Snow density (kg m-3).

    air_temp : ndarray(float, ndim=2)
        Air temperature (K).

    References
    ----------
    .. [1] Anderson, E. A. (1976). A point energy and mass balance model of a
       snow cover (NOAA Technical Report NWS 19, pp. 1–172). NOAA.
       https://repository.library.noaa.gov/view/noaa/6392

    .. [2] Koivusalo, H., Heikinheimo, M., & Karvonen, T. (2001). Test of a
       simple two-layer parameterisation to simulate the energy balance and
       temperature of a snow pack. Theoretical and Applied Climatology, 70(1–4),
       65–79. https://doi.org/10.1007/s007040170006
    """
    timestep_h = timestep / c.SECONDS_PER_HOUR  # model timestep (h)

    c1 = 0.001  # m2 kg-1 h-1
    c2 = 0.08  # K-1
    c3 = 0.021  # m3 kg-1
    c5 = 0.04  # K-1
    c7 = 0.046  # m3 kg-1
    rho_d = 150.  # kg m-3

    num_pixels = len(roi_idxs)
    for idx_num in prange(num_pixels):
        i, j = roi_idxs[idx_num]
        load = 0.  # snow load (mass of layers above + 50% of the current layer) (kg m-2)

        for k in range(num_layers[i, j]):
            if thickness[k, i, j] > 0:  # TODO is this necessary?
                cur_layer_mass = ice_content[k, i, j] + liquid_water_content[k, i, j]
                load += cur_layer_mass / 2.

                # TODO rather update density in update_layers()?
                density[k, i, j] = (ice_content[k, i, j] + liquid_water_content[k, i, j]) / thickness[k, i, j]

                if density[k, i, j] > rho_d:
                    c6 = np.exp(-c7 * (density[k, i, j] - rho_d))
                else:
                    c6 = 1.

                # Parameter c4 has an enhancement factor of 2 for wet snow
                if liquid_water_content[k, i, j] > 0:
                    c4 = 0.02  # h-1
                else:
                    c4 = 0.01  # h-1

                # Densification due to snow load (eq. (3.29) in [1], eq. (12) in [2])
                dens_compact = c1 * load * np.exp(-c2 * (c.T0 - air_temp[i, j]) - c3 * density[k, i, j])

                # Densification due to destructive metamorphism (eq. (3.31) in [1], eq. (13) in [2])
                dens_metamorph = c4 * np.exp(-c5 * (c.T0 - air_temp[i, j])) * c6

                densification_rate = density[k, i, j] * (dens_compact + dens_metamorph)
                if np.isnan(densification_rate):
                    densification_rate = 0.

                density[k, i, j] += densification_rate * timestep_h
                thickness[k, i, j] = (ice_content[k, i, j] + liquid_water_content[k, i, j]) / density[k, i, j]

                # Update snow load for the next layer
                load += cur_layer_mass / 2.


@njit(cache=True, parallel=True)
def _compaction_empirical(
    roi_idxs,
    timestep,
    timescale,
    max_cold_density,
    max_melting_density,
    num_layers,
    temp,
    thickness,
    ice_content,
    liquid_water_content,
    density,
):
    """
    Calculate snow compaction following [1].

    Parameters
    ----------
    roi_idxs : ndarray(int, ndim=2)
        (N, 2)-array specifying the (row, col) indices within the data arrays
        that should be considered.

    timestep : float
        Model timestep (s).

    timescale : float
        Snow compaction timescale (s).

    max_cold_density : float
        Maximum density for cold (T < 0 °C) snow (kg m-2).

    max_melting_density : float
        Maximum density for melting snow (kg m-2).

    num_layers : ndarray(float, ndim=2)
        Number of snow layers.

    temp : ndarray(float, ndim=3)
        Snow temperature (K).

    thickness : ndarray(float, ndim=3)
        Snow thickness (m).

    ice_content : ndarray(float, ndim=3)
        Ice content of snow (kg m-2).

    liquid_water_content : ndarray(float, ndim=3)
        Liquid water content of snow (kg m-2).

    density : ndarray(float, ndim=3)
        Snow density (kg m-3).

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    num_pixels = len(roi_idxs)
    for idx_num in prange(num_pixels):
        i, j = roi_idxs[idx_num]

        for k in range(num_layers[i, j]):
            if thickness[k, i, j] > 0:  # TODO is this necessary?
                # TODO rather update density in update_layers()?
                density[k, i, j] = (ice_content[k, i, j] + liquid_water_content[k, i, j]) / thickness[k, i, j]

                if temp[k, i, j] < c.T0:
                    max_density = max_cold_density
                else:
                    max_density = max_melting_density

                if density[k, i, j] < max_density:
                    # Where the maximum density is already reached, do not increase it anymore
                    # but to avoid jumps do not actually clip the values (because snow might
                    # switch between "cold" and "melting" from one timestep to another)
                    density[k, i, j] = (
                        max_density
                        + (density[k, i, j] - max_density)
                        * np.exp(-timestep / timescale)
                    )

                    thickness[k, i, j] = (ice_content[k, i, j] + liquid_water_content[k, i, j]) / density[k, i, j]


def snow_properties(model):
    """
    Update snow properties (depth, SWE, snow cover fraction, heat capacity).

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    snow = model.state.snow

    snow.depth[:] = snow.thickness.sum(axis=0)
    snow.swe[:] = (snow.ice_content + snow.liquid_water_content).sum(axis=0)

    if model.config.snow.snow_cover_fraction_depth_scale > 0:
        # Snow cover fraction (eq. (13))
        snow.area_fraction[:] = np.tanh(snow.depth[:] / model.config.snow.snow_cover_fraction_depth_scale)
    else:
        snow.area_fraction[:] = (snow.depth[:] > 0) * 1.

    # Areal heat capacity of snow (eq. (9))
    snow.heat_cap[:] = (
        snow.ice_content * c.SPEC_HEAT_CAP_ICE
        + snow.liquid_water_content * c.SPEC_HEAT_CAP_WATER
    )


def max_liquid_water_content(model):
    """
    Calculate the maximum liquid water content of snow.

    The following options for calculating the maximum liquid water content are
    supported:
        - 'pore_volume_fraction': maximum LWC is set to a fraction of the snow
          pore volume following [1].
        - 'mass_fraction': maximum LWC is set to a fraction of the snow
          ice mass.

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    Returns
    -------
    max_lwc : ndarray(float, ndim=3)
        Maximum liquid water content (kg m-2).

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    s = model.state
    method = model.config.snow.liquid_water_content.method

    if method == 'pore_volume_fraction':
        pos = s.snow.thickness > 0.
        porosity = np.zeros(s.snow.ice_content.shape)

        porosity[pos] = 1 - s.snow.ice_content[pos] / (c.ICE_DENSITY * s.snow.thickness[pos])  # eq. (27)

        max_lwc = (  # eq. (28)
            c.WATER_DENSITY
            * s.snow.thickness
            * porosity
            * model.config.snow.liquid_water_content.max
        )
    elif method == 'mass_fraction':
        max_lwc = model.config.snow.liquid_water_content.max * s.snow.ice_content

    return max_lwc



from loguru import logger
from openamundsen import modules
import numpy as np


class LandCover:
    def __init__(self, model):
        self.model = model

        s_lc = model.state.add_category('land_cover')
        s_lc.add_variable('land_cover', long_name='Land cover class', dtype=int, retain=True)
        s_lc.add_variable('plant_height', 'm', 'Plant height')
        s_lc.add_variable('lai', 'm2 m-2', 'Leaf area index')
        s_lc.add_variable('lai_eff', 'm2 m-2', 'Effective leaf area index')

    def initialize(self):
        model = self.model
        roi = model.grid.roi
        s = self.model.state

        # Prepare unique land cover classes occurring in the model domain and their associated pixel
        # locations
        lccs = np.unique(s.land_cover.land_cover[roi])
        lccs = lccs[lccs > 0]
        self.class_pixels = {}
        self.forest_classes = []  # LCCs with is_forest == True
        self.lai_classes = []  # LCCs with valid LAI parameters
        for lcc in lccs:
            try:
                lcc_params = model.config['land_cover']['classes'][lcc]
            except KeyError:
                logger.warning(f'Unknown land cover class: {lcc}')
                continue

            self.class_pixels[lcc] = (s.land_cover.land_cover == lcc) & roi

            if lcc_params.get('is_forest', False):
                self.forest_classes.append(lcc)

            if 'leaf_area_index' in lcc_params:
                if (
                    'min' in lcc_params['leaf_area_index']
                    and 'max' in lcc_params['leaf_area_index']
                ):
                    self.lai_classes.append(lcc)
                else:
                    logger.warning(f'Incomplete LAI parameters for land cover class {lcc}')

    def lai(self):
        model = self.model
        s_lc = self.model.state.land_cover

        for lcc in self.lai_classes:
            lcc_params = model.config['land_cover']['classes'][lcc]
            min_lai = lcc_params['leaf_area_index']['min']
            max_lai = lcc_params['leaf_area_index']['max']
            (
                length_ini,
                length_dev,
                length_mid,
                length_late,
            ) = self.growth_stage_lengths(lcc)
            growing_period_day = self.growing_period_day(lcc)

            lcc_lai = modules.evapotranspiration.crop_coefficient(
                growing_period_day,
                0,  # no initial period with constant LAI
                length_ini + length_dev,
                length_mid,
                length_late,
                min_lai,
                max_lai,
                min_lai,
                min_lai,
            )

            s_lc.lai[self.class_pixels[lcc]] = lcc_lai
            s_lc.lai_eff[self.class_pixels[lcc]] = (
                lcc_lai + lcc_params['leaf_area_index'].get('effective_add', 0)
            )

    def growth_stage_lengths(self, lcc):
        lcc_params = self.model.config['land_cover']['classes'][lcc]
        lcc_gsls = lcc_params['growth_stage_lengths']

        if isinstance(lcc_gsls[0], int):
            length_ini, length_dev, length_mid, length_late = lcc_gsls
        elif isinstance(lcc_gsls[0], list):  # growing season composed of several subcycles
            length_ini = [g[0] for g in lcc_gsls]
            length_dev = [g[1] for g in lcc_gsls]
            length_mid = [g[2] for g in lcc_gsls]
            length_late = [g[3] for g in lcc_gsls]

        return (length_ini, length_dev, length_mid, length_late)

    def growing_period_day(self, lcc):
        lcc_params = self.model.config['land_cover']['classes'][lcc]
        return self.model.date.dayofyear - lcc_params['plant_date'] + 1  # (1-based)



import openamundsen as oa
from pathlib import Path
import pooch
import pytest
import textwrap

version = oa.__version__
if '+' in version:
    data_url = 'https://github.com/openamundsen/testdata/raw/{version}/'
else:
    data_url = 'https://github.com/openamundsen/testdata/raw/v{version}/'

data_fetcher = pooch.create(
    path=pooch.os_cache('openamundsen-tests'),
    base_url=data_url,
    version=version,
    version_dev='main',
    registry={  # hashes generated with generate_hashes.py
        'grids/rofental/dem_rofental_1000.asc': '368ee374bfbb927bae5408a38dbafbbeab4ff59dc94ae30956b617bf28c114ec',
        'grids/rofental/lc_rofental_1000.asc': '5bde204bf9e79acdc47f9d9c70f21f20f669cfb7cbe65418ef42dd0a52676b98',
        'grids/rofental/roi_rofental_1000.asc': '5c495833e8465931d7aedc982ef093332572034d9cf7da5fc8b98b72111e3d25',
        'grids/rofental/soil_rofental_1000.asc': '48b46a8b3216a012393f36e34561473b3c1fd8dccef86d9a7e34aab9e3deb89b',
        'grids/rofental/srf_rofental_1000.asc': '431f0ed2181abc2a79c6ffd79b804e5c983c9c47003bce57b853db2b41abf62d',
        'grids/rofental/svf_rofental_1000.asc': 'a4b7704fad6f7c63445a76c64207ff881b73023ec7d958d4581cd7fb42d06966',
        'meteo/rofental/convert_csv_to_netcdf.py': 'f4e7aeb04d3eff7f1bdc6c4619600c4ab8dc560ffbc8feab83a30d200400f49f',
        'meteo/rofental/csv/bellavista.csv': '0c760f490aa7dc8d6af9da8987a6cf59807ec3e41c802a9c73b7773d269ec603',
        'meteo/rofental/csv/latschbloder.csv': 'f9e089bca0e7c931df51e1c624dc4985089c20c802d43de8b4c5175353a00931',
        'meteo/rofental/csv/proviantdepot.csv': '0e3a7eb07f9f127f0a257b17c62442e91860fb5a64f57387da547f69dcb61b6f',
        'meteo/rofental/csv/stations.csv': '175b420d9659d3478d4755fc4b8d8c2569400b37d84ffb9906e1ab661afe5dc4',
        'meteo/rofental/netcdf/bellavista.nc': '0fdeb860bce28a287bd48d2fc95f42a07ed842e9a0f0a0045f658605df2ff564',
        'meteo/rofental/netcdf/latschbloder.nc': 'cc96e0dbc4024b093fcc161d8f2c862bb8abd379a7d3dbd540d78242f17aef75',
        'meteo/rofental/netcdf/proviantdepot.nc': 'f8a276c98343c2574fb810eb4b3ee1b08bd29d9f5d0cfaa5a1104f1aa25af374',
        'results/canopy_point.nc': '3cd25b0119037be0eb95401ae7cfb618fc840d5244e4b2f42af45e72feff493c',
        'results/snow_cryolayers_point.nc': '04c6363c888bd6296e9ed3fa5325e418c659532b4f066d24fd33a34d5930877d',
        'results/snow_multilayer_point.nc': '702b7879a57f9837c97a629e66cd11fbb014d1b6e42f1241874de86d0714c380',
    },
)
DATA_DIR = data_fetcher.abspath

_BASE_CONFIG_YAML = textwrap.dedent(f'''
    domain: rofental
    start_date: 2020-01-15
    end_date: 2020-01-15
    resolution: 1000
    timestep: 3h
    crs: "epsg:32632"
    timezone: 1

    input_data:
      grids:
        dir: {DATA_DIR}/grids/rofental
      meteo:
        dir: {DATA_DIR}/meteo/rofental/netcdf

    output_data:
      timeseries:
        format: memory

      grids:
        format: memory
''')
_BASE_CONFIG = oa.parse_config(oa.Configuration.from_yaml(_BASE_CONFIG_YAML))


def pytest_addoption(parser):
    parser.addoption('--run-slow', action='store_true', default=False, help='run slow tests')
    parser.addoption(
        '--skip-comparisons',
        action='store_true',
        default=False,
        help='skip comparisons',
    )
    parser.addoption(
        '--prepare-comparison-data',
        action='store_true',
        default=False,
        help='prepare baseline model results for comparisons',
    )
    parser.addoption(
        '--comparison-data-dir',
        type=str,
        help='baseline model results directory',
    )
    parser.addoption(
        '--reports-dir',
        type=str,
        help='directory for writing comparison reports',
    )


def pytest_configure(config):
    config.addinivalue_line('markers', 'slow: marks test as slow')
    config.addinivalue_line('markers', 'comparison: marks test as comparing with baseline data')

    prepare_comparison_data = config.getoption('--prepare-comparison-data')

    comp_data_dir = config.getoption('--comparison-data-dir')
    if comp_data_dir is None:
        if prepare_comparison_data:
            raise Exception('--comparison-data-dir must be specified when '
                            '--prepare-comparison-data is set')
        else:
            comp_data_dir = DATA_DIR / 'results'
    if comp_data_dir is not None:
        comp_data_dir = Path(comp_data_dir)
        comp_data_dir.mkdir(parents=True, exist_ok=True)

    reports_dir = config.getoption('--reports-dir')
    if reports_dir is not None:
        reports_dir = Path(reports_dir)

        try:
            import plotly
        except ImportError:
            raise ImportError('plotly is required for report creation')

    pytest.DATA_DIR = DATA_DIR
    pytest.COMPARISON_DATA_DIR = comp_data_dir
    pytest.REPORTS_DIR = reports_dir
    pytest.PREPARE_COMPARISON_DATA = prepare_comparison_data


def pytest_collection_modifyitems(config, items):
    run_slow = config.getoption('--run-slow')
    skip_slow = pytest.mark.skip(reason='need --run-slow option to run')
    skip_comparisons = pytest.mark.skipif(
        config.getoption('--skip-comparisons'),
        reason='skipping comparisons',
    )

    for item in items:
        if 'slow' in item.keywords and not run_slow:
            item.add_marker(skip_slow)
        elif 'comparison' in item.keywords:
            item.add_marker(skip_comparisons)


def pytest_sessionstart(session):
    fetch_data_files()


def fetch_data_files():
    for file in data_fetcher.registry_files:
        data_fetcher.fetch(file)


def base_config():
    return _BASE_CONFIG.copy()



class SnowModel:
    def initialize(self):
        pass

    def albedo_aging(self):
        pass

    def compaction(self):
        pass

    def accumulation(self):
        pass

    def heat_conduction(self):
        pass

    def melt(self):
        pass

    def sublimation(self):
        pass

    def runoff(self):
        pass

    def update_layers(self):
        pass

    def update_properties(self):
        pass

    def add_snow(
            self,
            pos,
            ice_content,
            liquid_water_content=0,
            density=None,
            albedo=None,
    ):
        pass



from .conf import Configuration, parse_config, read_config
from .model import OpenAmundsen, Model
from . import constants, errors, terrain


# Get version (method as used by matplotlib: https://github.com/matplotlib/matplotlib/blob/bcc1ce8461f5b6e874baaaa02ef776d0243a4abe/lib/matplotlib/__init__.py#L133-L151)
def __getattr__(name):
    if name == '__version__':
        from pathlib import Path
        import setuptools_scm
        global __version__
        root = Path(__file__).resolve().parents[1]
        if (root / '.git').exists() and not (root / '.git/shallow').exists():
            __version__ = setuptools_scm.get_version(
                root=root,
                version_scheme='post-release',
                fallback_version='0.0.0+UNKNOWN',
            )
        else:
            try:
                from . import _version
                __version__ = _version.version
            except ImportError:
                __version__ = '0.0.0+UNKNOWN'
        return __version__
    raise AttributeError(f'module {__name__!r} has no attribute {name!r}')


__all__ = [
    'OpenAmundsen',
    'Configuration',
    'parse_config',
    'read_config',
]



import numpy as np
from numba import njit, prange
from openamundsen import constants, constants as c, heatconduction
from openamundsen.snowmodel import SnowModel
from . import snow


class MultilayerSnowModel(SnowModel):
    def __init__(self, model):
        self.model = model

        s = model.state.snow
        num_snow_layers = len(model.config.snow.min_thickness)

        s.add_variable('num_layers', '1', 'Number of snow layers', dtype=int, retain=True)
        s.add_variable('thickness', 'm', 'Snow thickness', dim3=num_snow_layers, retain=True)
        s.add_variable('density', 'kg m-3', 'Snow density', 'snow_density', dim3=num_snow_layers)
        s.add_variable('ice_content', 'kg m-2', 'Ice content of snow', dim3=num_snow_layers, retain=True)
        s.add_variable('liquid_water_content', 'kg m-2', 'Liquid water content of snow', 'liquid_water_content_of_snow_layer', dim3=num_snow_layers, retain=True)
        s.add_variable('temp', 'K', 'Snow temperature', dim3=num_snow_layers, retain=True)
        s.add_variable('therm_cond', 'W m-1 K-1', 'Thermal conductivity of snow', dim3=num_snow_layers, retain=True)
        s.add_variable('heat_cap', 'J K-1 m-2', 'Areal heat capacity of snow', dim3=num_snow_layers)

    def initialize(self):
        roi = self.model.grid.roi
        s = self.model.state.snow

        s.swe[roi] = 0
        s.depth[roi] = 0
        s.area_fraction[roi] = 0
        s.num_layers[roi] = 0
        s.sublimation[roi] = 0
        s.therm_cond[:, roi] = self.model.config.snow.thermal_conductivity
        s.thickness[:, roi] = 0
        s.ice_content[:, roi] = 0
        s.liquid_water_content[:, roi] = 0
        s.temp[:, roi] = constants.T0

    def albedo_aging(self):
        snow.albedo(self.model)

    def compaction(self):
        snow.compaction(self.model)

    def accumulation(self):
        model = self.model
        s = model.state
        pos = s.meteo.snowfall > 0
        self.add_snow(
            pos,
            s.meteo.snowfall[pos],
            density=snow.fresh_snow_density(s.meteo.wet_bulb_temp[pos]),
        )

    def heat_conduction(self):
        model = self.model
        s = model.state
        _heat_conduction(
            model.grid.roi_idxs,
            s.snow.num_layers,
            s.snow.thickness,
            s.soil.thickness,
            model.timestep,
            s.snow.temp,
            s.snow.therm_cond,
            s.soil.therm_cond,
            s.surface.heat_flux,
            s.snow.heat_cap,
        )

    def melt(self):
        model = self.model
        s = model.state
        _melt(
            model.grid.roi_idxs,
            model.timestep,
            s.snow.num_layers,
            s.snow.melt,
            s.snow.thickness,
            s.snow.temp,
            s.snow.ice_content,
            s.snow.liquid_water_content,
            s.snow.heat_cap,
        )

    def sublimation(self):
        model = self.model
        s = model.state
        roi = model.grid.roi

        s.snow.sublimation[roi] = 0.
        pos_roi = (s.snow.ice_content[0, roi] > 0) | (s.surface.temp[roi] < constants.T0)
        pos = model.roi_mask_to_global(pos_roi)
        pot_sublim = -1 * s.surface.moisture_flux[pos] * model.timestep
        pot_sublim[np.isnan(pot_sublim)] = 0.
        s.snow.sublimation[pos] = np.minimum(pot_sublim, s.snow.ice_content[:, pos].sum(axis=0))

        # First resublimation
        frost = -np.minimum(s.snow.sublimation, 0)
        pos = frost > 0
        self.add_snow(
            pos,
            frost[pos],
            density=snow.fresh_snow_density(s.meteo.wet_bulb_temp[pos]),
        )

        # Then sublimation
        _sublimation(
            model.grid.roi_idxs,
            model.timestep,
            s.snow.num_layers,
            s.snow.ice_content,
            s.snow.thickness,
            s.snow.sublimation,
        )

    def runoff(self):
        model = self.model
        s = model.state
        _runoff(
            model.grid.roi_idxs,
            snow.max_liquid_water_content(model),
            s.meteo.rainfall,
            s.snow.num_layers,
            s.snow.thickness,
            s.snow.temp,
            s.snow.ice_content,
            s.snow.liquid_water_content,
            s.snow.refreezing,
            s.snow.runoff,
            s.snow.heat_cap,
        )

    def update_layers(self):
        model = self.model
        s = model.state
        _update_layers(
            model.grid.roi_idxs,
            s.snow.num_layers,
            np.array(model.config.snow.min_thickness),
            s.snow.thickness,
            s.snow.ice_content,
            s.snow.liquid_water_content,
            s.snow.heat_cap,
            s.snow.temp,
            s.snow.density,
            s.snow.depth,
        )
        s.snow.albedo[s.snow.num_layers == 0] = np.nan

    def update_properties(self):
        snow.snow_properties(self.model)

    def add_snow(
            self,
            pos,
            ice_content,
            liquid_water_content=0,
            density=None,
            albedo=None,
    ):
        """
        Add snow to the top of the snowpack.
        """
        model = self.model
        s = model.state

        ice_content = np.nan_to_num(ice_content, nan=0., copy=True)

        pos_init = (s.snow.num_layers[pos] == 0) & (ice_content > 0)
        pos_init_global = model.global_mask(pos_init, pos)

        # If albedo is None, set it to the maximum albedo for currently snow-free pixels and keep
        # the current albedo for the other pixels
        if albedo is None:
            albedo = s.snow.albedo[pos]
            albedo[pos_init] = model.config.snow.albedo.max

        s.snow.albedo[pos] = albedo

        # Initialize first snow layer where necessary
        s.snow.num_layers[pos_init_global] = 1
        s.snow.temp[0, pos_init_global] = np.minimum(s.meteo.temp[pos_init_global], constants.T0)

        # Add snow to first layer
        s.snow.ice_content[0, pos] += ice_content
        s.snow.liquid_water_content[0, pos] += liquid_water_content
        s.snow.thickness[0, pos] += ice_content / density


@njit(cache=True, parallel=True)
def _melt(
    roi_idxs,
    timestep,
    num_layers,
    melt,
    thickness,
    temp,
    ice_content,
    liquid_water_content,
    heat_cap,
):
    """
    Calculate snowmelt following [1].

    Parameters
    ----------
    roi_idxs : ndarray(int, ndim=2)
        (N, 2)-array specifying the (row, col) indices within the data arrays
        that should be considered.

    timestep : float
        Model timestep (s).

    num_layers : ndarray(float, ndim=2)
        Number of snow layers.

    melt : ndarray(float, ndim=2)
        Snowmelt (kg m-2).

    thickness : ndarray(float, ndim=3)
        Snow thickness (m).

    temp : ndarray(float, ndim=3)
        Snow temperature (K).

    ice_content : ndarray(float, ndim=3)
        Ice content of snow (kg m-2).

    liquid_water_content : ndarray(float, ndim=3)
        Liquid water content of snow (kg m-2).

    heat_cap : ndarray(float, ndim=3)
        Areal heat capacity of snow (J K-1 m-2).

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    num_pixels = len(roi_idxs)
    for idx_num in prange(num_pixels):
        i, j = roi_idxs[idx_num]

        ice_content_change = melt[i, j]
        actual_melt = 0.  # actual melt without cold content reduction

        for k in range(num_layers[i, j]):
            cold_content = heat_cap[k, i, j] * (c.T0 - temp[k, i, j])
            if cold_content < 0:
                ice_content_change -= cold_content / c.LATENT_HEAT_OF_FUSION
                temp[k, i, j] = c.T0

            if ice_content_change > 0:
                if ice_content_change > ice_content[k, i, j]:  # layer melts completely
                    ice_content_change -= ice_content[k, i, j]
                    thickness[k, i, j] = 0.
                    liquid_water_content[k, i, j] += ice_content[k, i, j]
                    actual_melt += ice_content[k, i, j]
                    ice_content[k, i, j] = 0.
                else:  # layer melts partially
                    thickness[k, i, j] *= (1 - ice_content_change / ice_content[k, i, j])
                    ice_content[k, i, j] -= ice_content_change
                    liquid_water_content[k, i, j] += ice_content_change
                    actual_melt += ice_content_change
                    ice_content_change = 0.

        melt[i, j] = actual_melt


@njit(cache=True, parallel=True)
def _sublimation(
    roi_idxs,
    timestep,
    num_layers,
    ice_content,
    thickness,
    sublimation,
):
    """
    Calculate snow sublimation following [1].

    Parameters
    ----------
    roi_idxs : ndarray(int, ndim=2)
        (N, 2)-array specifying the (row, col) indices within the data arrays
        that should be considered.

    timestep : float
        Model timestep (s).

    num_layers : ndarray(float, ndim=2)
        Number of snow layers.

    ice_content : ndarray(float, ndim=3)
        Ice content of snow (kg m-2).

    thickness : ndarray(float, ndim=3)
        Snow thickness (m).

    sublimation : ndarray(float, ndim=2)
        Snow sublimation (kg m-2).

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    num_pixels = len(roi_idxs)
    for idx_num in prange(num_pixels):
        i, j = roi_idxs[idx_num]

        ice_content_change = max(sublimation[i, j], 0.)

        if ice_content_change > 0:
            for k in range(num_layers[i, j]):
                if ice_content_change > ice_content[k, i, j]:  # complete sublimation of layer
                    ice_content_change -= ice_content[k, i, j]
                    thickness[k, i, j] = 0.
                    ice_content[k, i, j] = 0.
                else:  # partial sublimation
                    thickness[k, i, j] *= (1 - ice_content_change / ice_content[k, i, j])
                    ice_content[k, i, j] -= ice_content_change
                    ice_content_change = 0.


@njit(cache=True, parallel=True)
def _runoff(
    roi_idxs,
    max_liquid_water_content,
    rainfall,
    num_layers,
    thickness,
    temp,
    ice_content,
    liquid_water_content,
    refreezing,
    runoff,
    heat_cap,
):
    """
    Calculate snowmelt runoff following [1].

    Parameters
    ----------
    roi_idxs : ndarray(int, ndim=2)
        (N, 2)-array specifying the (row, col) indices within the data arrays
        that should be considered.

    max_liquid_water_content : ndarray(float, ndim=3)
        Maximum liquid water content (kg m-2).

    rainfall : ndarray(float, ndim=2)
        Rainfall amount (kg m-2).

    num_layers : ndarray(float, ndim=2)
        Number of snow layers.

    thickness : ndarray(float, ndim=3)
        Snow thickness (m).

    temp : ndarray(float, ndim=3)
        Snow temperature (K).

    ice_content : ndarray(float, ndim=3)
        Ice content of snow (kg m-2).

    liquid_water_content : ndarray(float, ndim=3)
        Liquid water content of snow (kg m-2).

    refreezing : ndarray(float, ndim=2)
        Liquid water refreezing (kg m-2).

    runoff : ndarray(float, ndim=2)
        Snow runoff (kg m-2).

    heat_cap : ndarray(float, ndim=3)
        Areal heat capacity of snow (J K-1 m-2).

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    num_pixels = len(roi_idxs)
    for idx_num in prange(num_pixels):
        i, j = roi_idxs[idx_num]

        refreezing[i, j] = 0.

        runoff[i, j] = rainfall[i, j]
        if np.isnan(runoff[i, j]):
            runoff[i, j] = 0.

        for k in range(num_layers[i, j]):
            liquid_water_content[k, i, j] += runoff[i, j]

            if liquid_water_content[k, i, j] > max_liquid_water_content[k, i, j]:
                runoff[i, j] = liquid_water_content[k, i, j] - max_liquid_water_content[k, i, j]
                liquid_water_content[k, i, j] = max_liquid_water_content[k, i, j]
            else:
                runoff[i, j] = 0.

            # Refreeze liquid water
            cold_content = heat_cap[k, i, j] * (c.T0 - temp[k, i, j])
            if cold_content > 0:
                ice_content_change = min(
                    liquid_water_content[k, i, j],
                    cold_content / c.LATENT_HEAT_OF_FUSION,
                )
                liquid_water_content[k, i, j] -= ice_content_change
                ice_content[k, i, j] += ice_content_change
                refreezing[i, j] += ice_content_change
                temp[k, i, j] += c.LATENT_HEAT_OF_FUSION * ice_content_change / heat_cap[k, i, j]


@njit(parallel=True, cache=True)
def _heat_conduction(
    roi_idxs,
    num_layers,
    snow_thickness,
    soil_thickness,
    timestep,
    temp,
    therm_cond_snow,
    therm_cond_soil,
    heat_flux,
    heat_cap,
):
    """
    Update snow layer temperatures.

    Parameters
    ----------
    roi_idxs : ndarray(int, ndim=2)
        (N, 2)-array specifying the (row, col) indices within the data arrays
        that should be considered.

    num_layers : ndarray(float, ndim=2)
        Number of snow layers.

    snow_thickness : ndarray(float, ndim=3)
        Snow thickness (m).

    soil_thickness : ndarray(float, ndim=3)
        Soil thickness (m).

    timestep : float
        Model timestep (s).

    temp : ndarray(float, ndim=3)
        Snow temperature (K).

    therm_cond_snow : ndarray(float, ndim=3)
        Snow thermal conductivity (W m-1 K-1).

    therm_cond_soil : ndarray(float, ndim=3)
        Soil thermal conductivity (W m-1 K-1).

    heat_flux : ndarray(float, ndim=2)
        Surface heat flux (W m-2).

    heat_cap : ndarray(float, ndim=3)
        Areal heat capacity of snow (J K-1 m-2).

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    num_pixels = len(roi_idxs)
    for idx_num in prange(num_pixels):
        i, j = roi_idxs[idx_num]

        ns = num_layers[i, j]

        if ns > 0:
            temp[:ns, i, j] += heatconduction.temp_change(
                snow_thickness[:ns, i, j],
                timestep,
                temp[:ns, i, j],
                therm_cond_snow[:ns, i, j],
                temp[-1, i, j],
                soil_thickness[0, i, j],
                therm_cond_soil[0, i, j],
                heat_flux[i, j],
                heat_cap[:ns, i, j],
            )


@njit(cache=True, parallel=True)
def _update_layers(
    roi_idxs,
    num_layers,
    min_thickness,
    thickness,
    ice_content,
    liquid_water_content,
    heat_cap,
    temp,
    density,
    depth,
):
    """
    Update snow layers.

    Parameters
    ----------
    roi_idxs : ndarray(int, ndim=2)
        (N, 2)-array specifying the (row, col) indices within the data arrays
        that should be considered.

    num_layers : ndarray(float, ndim=2)
        Number of snow layers.

    min_thickness : ndarray(float, ndim=1)
        Minimum snow layer thicknesses (m).

    thickness : ndarray(float, ndim=3)
        Snow thickness (m).

    ice_content : ndarray(float, ndim=3)
        Ice content of snow (kg m-2).

    liquid_water_content : ndarray(float, ndim=3)
        Liquid water content of snow (kg m-2).

    heat_cap : ndarray(float, ndim=3)
        Areal heat capacity of snow (J K-1 m-2).

    temp : ndarray(float, ndim=3)
        Snow temperature (K).

    density : ndarray(float, ndim=3)
        Snow density (kg m-3).

    depth : ndarray(float, ndim=2)
        Snow depth (m).

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    max_num_layers = len(min_thickness)
    num_layers_prev = num_layers.copy()
    thickness_prev = thickness.copy()
    ice_content_prev = ice_content.copy()
    liquid_water_content_prev = liquid_water_content.copy()
    energy_prev = heat_cap * (temp - c.T0)  # energy content (J m-2)

    num_pixels = len(roi_idxs)
    for idx_num in prange(num_pixels):
        i, j = roi_idxs[idx_num]

        num_layers[i, j] = 0
        thickness[:, i, j] = 0.
        ice_content[:, i, j] = 0.
        liquid_water_content[:, i, j] = 0.
        temp[:, i, j] = c.T0
        density[:, i, j] = np.nan
        internal_energy = np.zeros(max_num_layers)

        if depth[i, j] > 0:
            new_thickness = depth[i, j]

            # Update thicknesses and number of layers
            for k in range(max_num_layers):
                thickness[k, i, j] = min_thickness[k]
                new_thickness -= min_thickness[k]

                if new_thickness <= min_thickness[k] or k == max_num_layers - 1:
                    thickness[k, i, j] += new_thickness
                    break

            # Set thin snow layers to 0 to avoid numerical artifacts
            # TODO should this be done at some other location?
            for k in range(max_num_layers):
                if thickness[k, i, j] < 1e-6:
                    thickness[k, i, j] = 0.

            ns = (thickness[:, i, j] > 0).sum()  # new number of layers
            new_thickness = thickness[0, i, j]
            k_new = 0

            # TODO optimize this loop
            for k_old in range(num_layers_prev[i, j]):
                while True:  # TODO replace with normal loop
                    weight = min(new_thickness / thickness_prev[k_old, i, j], 1.)

                    ice_content[k_new, i, j] += weight * ice_content_prev[k_old, i, j]
                    liquid_water_content[k_new, i, j] += weight * liquid_water_content_prev[k_old, i, j]
                    internal_energy[k_new] += weight * energy_prev[k_old, i, j]

                    if weight == 1.:
                        new_thickness -= thickness_prev[k_old, i, j]
                        break

                    thickness_prev[k_old, i, j] *= 1 - weight
                    ice_content_prev[k_old, i, j] *= 1 - weight
                    liquid_water_content_prev[k_old, i, j] *= 1 - weight
                    energy_prev[k_old, i, j] *= 1 - weight

                    k_new += 1

                    if k_new >= ns:
                        break

                    if weight < 1:
                        new_thickness = thickness[k_new, i, j]

            num_layers[i, j] = ns

            # Update areal heat capacity and snow temperature
            heat_cap[:ns, i, j] = (  # TODO use snow_heat_capacity() for this
                ice_content[:ns, i, j] * c.SPEC_HEAT_CAP_ICE
                + liquid_water_content[:ns, i, j] * c.SPEC_HEAT_CAP_WATER
            )
            temp[:ns, i, j] = c.T0 + internal_energy[:ns] / heat_cap[:ns, i, j]

            # Update density
            density[:ns, i, j] = (
                (liquid_water_content[:ns, i, j] + ice_content[:ns, i, j])
                / thickness[:ns, i, j]
            )



from dataclasses import dataclass
from munch import Munch
import numpy as np
from openamundsen import constants
import pandas as pd
import pandas.tseries.frequencies
from pathlib import Path, PosixPath, WindowsPath
import pyproj
import rasterio
import ruamel.yaml
from typing import Union


class ConfigurationYAML(ruamel.yaml.YAML):
    def __init__(self):
        super().__init__(typ='rt')  # .indent() works only with the roundtrip dumper
        self.default_flow_style = False
        self.indent(mapping=2, sequence=4, offset=2)

        self.representer.add_representer(pd.Timestamp, self._repr_datetime)

        # Add representers for path objects (just using pathlib.Path does not
        # work, so have to add PosixPath and WindowsPath separately)
        self.representer.add_representer(PosixPath, self._repr_path)
        self.representer.add_representer(WindowsPath, self._repr_path)

    def _repr_datetime(self, representer, date):
        return representer.represent_str(str(date))

    def _repr_path(self, representer, path):
        return representer.represent_str(str(path))

    def dump(self, data, stream=None, **kw):
        inefficient = False

        if stream is None:
            inefficient = True
            stream = ruamel.yaml.compat.StringIO()

        super().dump(data, stream, **kw)

        if inefficient:
            return stream.getvalue()


yaml = ConfigurationYAML()


def read_yaml_file(filename):
    """
    Read a YAML file.

    Parameters
    ----------
    filename : str

    Returns
    -------
    result : dict
    """
    with open(filename) as f:
        return load_yaml(f.read())


def load_yaml(s):
    return yaml.load(s)


def to_yaml(d):
    return yaml.dump(d)


def raster_filename(kind, config):
    """
    Return the filename of an input raster file for a model run.

    Parameters
    ----------
    kind : str
        Type of input file, e.g. 'dem' or 'roi'.

    config : dict
        Model run configuration.

    Returns
    -------
    file : pathlib.Path
    """
    dir = config['input_data']['grids']['dir']
    domain = config['domain']
    resolution = config['resolution']
    extension = 'asc'
    return Path(f'{dir}/{kind}_{domain}_{resolution}.{extension}')


def transform_coords(x, y, src_crs, dst_crs):
    """
    Transform coordinates from one coordinate reference system to another.

    Parameters
    ----------
    x : ndarray
        Source x coordinates or longitude.

    y : ndarray
        Source y coordinates or latitude.

    src_crs
        CRS of the original coordinates as a pyproj-compatible input
        (e.g. a string "epsg:xxxx").

    dst_crs
        Target CRS.

    Returns
    -------
    (x, y) tuple of ndarrays containing the transformed coordinates.
    """
    x = np.array(x)
    y = np.array(y)
    transformer = pyproj.Transformer.from_crs(src_crs, dst_crs, always_xy=True)
    return transformer.transform(x, y)


class ModelGrid(Munch):
    """
    Container for storing model grid related variables.
    """
    def prepare_coordinates(self):
        """
        Prepare a range of variables related to the grid coordinates:
        - xs, ys: 1d-arrays containing the x and y coordinates in the grid CRS.
        - X, Y, 2d-arrays containing the x and y coordinates for each grid point.
        - all_points: (N, 2)-array containing (x, y) coordinates of all grid points.
        - roi_points: (N, 2)-array containing (x, y) coordinates of all ROI points.
        - roi_idxs: (N, 2)-array containing (row, col) indexes of all ROI points.
        - roi_idxs_flat: 1d-array containing the flattened (1d) indexes of all ROI points
        """
        transform = self.transform
        if transform.a < 0 or transform.e > 0:
            raise NotImplementedError  # we only allow left-right and top-down oriented grids

        x_min = transform.xoff
        x_max = x_min + self.cols * self.resolution
        y_max = transform.yoff
        y_min = y_max - self.rows * self.resolution

        x_range, y_range = rasterio.transform.xy(
            self.transform,
            [0, self.rows - 1],
            [0, self.cols - 1],
        )
        xs = np.linspace(x_range[0], x_range[1], self.cols)
        ys = np.linspace(y_range[0], y_range[1], self.rows)
        X, Y = np.meshgrid(xs, ys)

        center_x = xs.mean()
        center_y = ys.mean()
        center_lon, center_lat = transform_coords(
            center_x,
            center_y,
            self.crs,
            constants.CRS_WGS84,
        )

        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max
        self.xs = xs
        self.ys = ys
        self.X = X
        self.Y = Y
        self.all_points = np.column_stack((X.flat, Y.flat))
        self.center_lon = center_lon
        self.center_lat = center_lat

        self.extended_grid = Munch(dict(
            available=False,
            rows=None,
            cols=None,
            row_offset=None,
            col_offset=None,
            row_slice=None,
            col_slice=None,
            dem=None,
            svf=None,
            normal_vec=None,
            shadows=None,
        ))
        # ("shadows" required because the shadows are calculated in
        # clear_sky_shortwave_irradiance(), but are required again later in shortwave_irradiance()
        # for calculating the clear sky irradiance for the extended-grid stations)

        self.prepare_roi_coordinates()

    def prepare_roi_coordinates(self):
        """
        Update the roi_points and roi_idxs variables using the ROI field.
        """
        if 'roi' not in self:
            self.roi = np.ones((self.rows, self.cols), dtype=bool)

        roi_xs = self.X[self.roi]
        roi_ys = self.Y[self.roi]
        self.roi_points = np.column_stack((roi_xs, roi_ys))
        self.roi_idxs = np.array(np.where(self.roi)).T
        self.roi_idxs_flat = np.where(self.roi.flat)[0]


def to_offset(offset: Union[str, pd.offsets.BaseOffset]) -> pd.offsets.BaseOffset:
    """
    Convert a pandas-compatible offset (e.g. '3h') to a DateOffset object.
    """
    # As of pandas 2.2.0, some frequency aliases have been deprecated (see
    # https://pandas.pydata.org/docs/dev/whatsnew/v2.2.0.html). In order to avoid deprecation
    # warnings in case the old aliases are still used in openAMUNDSEN configurations, we replace
    # the deprecated aliases "H", "M" and "Y" here with their new versions.
    # Furthermore, we parse the offsets "ME" and "YE" (which have been introduced in pandas 2.2.0)
    # here manually in order to make them work also with earlier pandas versions.
    if isinstance(offset, str):
        if offset.endswith('H'):
            offset = offset[:-1] + 'h'
        elif offset.endswith('M'):
            offset = offset[:-1] + 'ME'
        elif offset.endswith('Y'):
            offset = offset[:-1] + 'YE'

        if offset == 'ME':
            offset = pd.offsets.MonthEnd()
        elif offset == 'YE':
            offset = pd.offsets.YearEnd()

    return pandas.tseries.frequencies.to_offset(offset)


def offset_to_timedelta(offset: Union[str, pd.offsets.BaseOffset]) -> pd.Timedelta:
    """
    Convert a pandas-compatible offset (e.g. '3h') to a Timedelta object.
    """
    return pd.to_timedelta(to_offset(offset))


@dataclass(frozen=True)
class TimestepProperties:
    first_of_run: bool
    strict_first_of_year: bool
    strict_first_of_month: bool
    strict_first_of_day: bool
    first_of_year: bool
    first_of_month: bool
    first_of_day: bool
    last_of_run: bool
    strict_last_of_year: bool
    strict_last_of_month: bool
    strict_last_of_day: bool
    last_of_year: bool
    last_of_month: bool
    last_of_day: bool


def normalize_array(data, min, max):
    """
    Normalize an array within a range.

    Parameters
    ----------
    data : array-like
        Data values.

    min : numeric
       Minimum of the normalized array.

    max : numeric
       Maximum of the normalized array.

    Returns
    -------
    data_norm : np.array
        Data normalized within [min, max].
    """
    data = np.asarray(data)
    data_min = np.nanmin(data)
    data_max = np.nanmax(data)
    scale_factor = (max - min) / (data_max - data_min)
    return min + scale_factor * (data - data_min)



from copy import deepcopy
import openamundsen as oa
import openamundsen.errors as errors
import pandas as pd
import pytest
from ruamel.yaml import YAML
import tempfile
from textwrap import dedent


def save_and_read_yaml_dict(d):
    yaml = YAML()

    with tempfile.TemporaryDirectory() as temp_dir:
        filename = f'{temp_dir}/config.yml'
        with open(filename, 'w') as f:
            yaml.dump(d, f)

        return oa.read_config(filename)


def parse_yaml_config(config):
    return oa.parse_config(oa.Configuration.from_yaml(config))


def save_and_parse_config(config):
    with tempfile.TemporaryDirectory() as temp_dir:
        filename = f'{temp_dir}/config.yml'

        with open(filename, 'w') as f:
            if isinstance(config, dict):
                yaml = YAML()
                yaml.dump(config, f)
            elif isinstance(config, str):
                f.write(config)
                f.flush()
            else:
                raise Exception('Unsupported type')

        return oa.parse_config(oa.read_config(filename))


@pytest.fixture(scope='function')
def minimal_config():
    return {
        'domain': 'dummy',
        'start_date': '2019-11-01',
        'end_date': '2020-04-30',
        'resolution': 50,
        'timezone': 1,
    }


def test_config_equal():
    config = {
        'domain': 'dummy',
        'start_date': '2019-11-01',
        'end_date': '2020-04-30',
        'resolution': 50,
        'timezone': 1,

        'input_data': {
            'grids': {
                'dir': 'dummydir',
            },

            'meteo': {
                'dir': 'meteodummydir',
                'format': 'netcdf',
            },
        },
    }

    config_read = save_and_read_yaml_dict(config)

    assert config == config_read


def test_read_config(minimal_config):
    config = save_and_parse_config(minimal_config)
    assert config['end_date'].hour == 23

    yaml_str = dedent("""
        domain: dummy
        start_date: 2019-11-01
        end_date: 2020-12-31
        resolution: 50
        timezone: 1
    """)
    config = parse_yaml_config(yaml_str)
    assert config['end_date'].hour == 23


def test_start_end_date_order(minimal_config):
    mc = deepcopy(minimal_config)
    mc['start_date'] = '2020-01-01'
    mc['end_date'] = '2020-01-01'
    oa.parse_config(mc)
    mc['timestep'] = 'D'
    oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc['start_date'] = '2020-01-01 00:00'
    mc['end_date'] = '2020-01-01 00:00'
    oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc['start_date'] = '2020-01-01'
    mc['end_date'] = '2019-12-31'
    with pytest.raises(errors.ConfigurationError):
        oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc['start_date'] = '2020-01-01 01:00'
    mc['end_date'] = '2020-01-01 00:00'
    with pytest.raises(errors.ConfigurationError):
        oa.parse_config(mc)


def test_infer_end_date(minimal_config):
    config = oa.parse_config(minimal_config)
    assert config['end_date'].hour == 23

    mc = deepcopy(minimal_config)
    mc['end_date'] = '2020-04-30 11:00'
    config = oa.parse_config(mc)
    assert config['end_date'].hour == 11

    mc = deepcopy(minimal_config)
    mc['end_date'] = ' 2020-04-30 '
    config = oa.parse_config(mc)
    assert config['end_date'].hour == 23

    mc = deepcopy(minimal_config)
    mc['timestep'] = 'D'
    config = oa.parse_config(mc)
    assert config['end_date'].hour == 0

    mc = deepcopy(minimal_config)
    mc['start_date'] = '2017-11-01'
    mc['end_date'] = '2018-04-25'
    mc['timestep'] = '5D'
    mc['output_data'] = {
        'timeseries': {
            'write_freq': '30D',
        },
    }
    config = oa.parse_config(mc)
    assert config['end_date'] == pd.Timestamp(mc['end_date'])

    yaml_str = dedent("""
        domain: dummy
        start_date: 2019-11-01
        end_date: {d}
        resolution: 50
        timezone: 1
    """)

    config = parse_yaml_config(yaml_str.format(d='2020-04-30'))
    assert config['end_date'].hour == 23

    config = parse_yaml_config(yaml_str.format(d='"2020-04-30"'))
    assert config['end_date'].hour == 23

    config = parse_yaml_config(yaml_str.format(d='2020-04-30 11:00'))
    assert config['end_date'].hour == 11

    config = parse_yaml_config(yaml_str.format(d='"2020-04-30 11:00"'))
    assert config['end_date'].hour == 11


def test_timesteps(minimal_config):
    mc = deepcopy(minimal_config)
    mc['start_date'] = '2019-11-01 03:00'
    oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc['start_date'] = '2019-11-01 00:15'
    mc['end_date'] = '2019-11-01 23:45'
    mc['timestep'] = '15min'
    oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc['start_date'] = '2019-11-01 02:00'
    mc['end_date'] = '2019-11-01 22:00'
    mc['timestep'] = '3h'
    with pytest.raises(errors.ConfigurationError):
        oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc['end_date'] = '2019-12-31 15:00'
    mc['timestep'] = 'D'
    with pytest.raises(errors.ConfigurationError):
        oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc['start_date'] = '2019-12-31 01:23'
    mc['end_date'] = '2019-12-31 03:45'
    mc['timestep'] = 'h'
    with pytest.raises(errors.ConfigurationError):
        oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc['start_date'] = '2017-11-01'
    mc['end_date'] = '2018-04-29'
    mc['timestep'] = '5D'
    with pytest.raises(errors.ConfigurationError):
        oa.parse_config(mc)


def test_missing_parameter(minimal_config):
    for key in minimal_config.keys():
        config = deepcopy(minimal_config)
        del config[key]

        with pytest.raises(errors.ConfigurationError):
            oa.parse_config(config)


def test_parse_twice(minimal_config):
    c1 = oa.parse_config(minimal_config)
    c2 = oa.parse_config(c1)
    assert c1 == c2



import numpy as np
import plotly.graph_objs as go
from plotly.subplots import make_subplots


def plot_point_comparison(
    ds_base,
    ds_dev,
    plot_vars,
    changed_vars,
    num_cols=3,
    plot_height=400,
    base_color='#1f77b4',
    dev_color='#d62728',
):
    num_rows = int(np.ceil(len(plot_vars) / num_cols))

    fig = make_subplots(
        rows=num_rows,
        cols=num_cols,
        subplot_titles=plot_vars,
        shared_xaxes=True,
    )

    highlight_subplots = []

    for plot_num, v in enumerate(plot_vars):
        data_base = ds_base[v].to_pandas()
        data_dev = ds_dev[v].to_pandas()

        row = plot_num // num_cols + 1
        col = plot_num % num_cols + 1

        num_dims = len(ds_base[v].dims)

        if num_dims == 1:
            data_base = data_base.to_frame()
            data_dev = data_dev.to_frame()

        for dim in range(num_dims):
            fig.add_trace(
                go.Scatter(
                    x=data_base.index,
                    y=data_base[dim],
                    name='base',
                    mode='lines',
                    line=dict(color=base_color),
                    showlegend=False,
                ),
                row=row,
                col=col,
            )

        if v in changed_vars:
            highlight_subplots.append(plot_num)

            for dim in range(num_dims):
                fig.add_trace(
                    go.Scatter(
                        x=data_dev.index,
                        y=data_dev[dim],
                        name='dev',
                        mode='lines',
                        line=dict(color=dev_color),
                        showlegend=False,
                    ),
                    row=row,
                    col=col,
                )

    for plot_num in highlight_subplots:
        fig['layout']['annotations'][plot_num]['font']['color'] = 'red'

    fig.update_layout(height=plot_height * num_rows)

    return fig


def plot_gridded_comparison(
    ds_base,
    ds_dev,
    plot_vars,
    changed_vars,
    plot_height=400,
):
    titles = []
    num_rows = 0
    for v in plot_vars:
        time_dim = ds_base[v].dims[0]
        dates = ds_base[v].coords[time_dim].to_index()
        num_rows += len(dates)
        for d in dates:
            titles.append(f'{v} ({d})')
            titles.append('')  # right column

    fig = make_subplots(
        rows=num_rows,
        cols=2,
        subplot_titles=titles,
        shared_xaxes=True,
        # shared_yaxes=True,
    )

    row = 1
    for v in plot_vars:
        data_base = ds_base[v].values
        data_dev = ds_dev[v].values

        for var_plot_num in range(data_base.shape[0]):
            data_base_cur = data_base[var_plot_num, :, :]
            data_dev_cur = data_dev[var_plot_num, :, :]

            min_val = min(
                np.nanmin(data_base_cur),
                np.nanmin(data_dev_cur),
            )
            max_val = max(
                np.nanmax(data_base_cur),
                np.nanmax(data_dev_cur),
            )

            fig.add_trace(
                go.Heatmap(
                    z=data_base_cur,
                    zmin=min_val,
                    zmax=max_val,
                    x=ds_base.x,
                    y=ds_base.y,
                    colorscale='viridis',
                    showscale=False,
                ),
                row=row,
                col=1,
            )

            if v in changed_vars and var_plot_num in changed_vars[v]:
                fig.add_trace(
                    go.Heatmap(
                        z=data_dev_cur,
                        zmin=min_val,
                        zmax=max_val,
                        x=ds_base.x,
                        y=ds_base.y,
                        colorscale='viridis',
                        showscale=False,
                    ),
                    row=row,
                    col=2,
                )

            row += 1

    fig.for_each_yaxis(lambda ax: ax.update(scaleanchor=ax.anchor))
    fig.update_layout(height=plot_height * num_rows)

    return fig


def fig_to_html(fig, filename, create_dir=True):
    if create_dir:
        filename.parent.mkdir(parents=True, exist_ok=True)

    with open(filename, 'w') as f:
        f.write(fig.to_html(include_plotlyjs='cdn'))



from loguru import logger
import numpy as np
from openamundsen import constants, interpolation, meteo


# Transfer function coefficents for 10 m wind speeds from [2] (Table 3).
# Values in the tuples correspond to (a, b, c, max_U).
KOCHENDORFER_COEFFS = {
    'us_un': (0.045, 1.21, 0.66, 8),
    'all_sa': (0.03, 1.04, 0.57, 12),
    'nor_sa': (0.05, 0.66, 0.23, 12),
    'us_sa': (0.03, 1.06, 0.63, 12),
    'us_da': (0.021, 0.74, 0.66, 8),
    'us_bda': (0.01, 0.48, 0.51, 8),
    'us_sdfir': (0.004, 0.00, 0.00, 8),
}


def correct_station_precipitation(model):
    """
    Correct station precipitation recordings for wind-induced undercatch.

    Precipitation can be corrected using the following `method`s in the model
    configuration (config['meteo']['precipitation_correction']):

        - 'constant_scf': Snowfall is corrected using a constant snow correction
          factor (scaled accordingly for mixed precipitation).
        - 'wmo': Snowfall and mixed precipitation are corrected using the
          transfer functions derived for different gauges from [1].
        - 'kochendorfer': Precipitation (regardless of phase) is corrected using
          the transfer frunctions derived for different gauges from [2].

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    References
    ----------
    .. [1] Goodison, B. E., Louie, P., & Yang, D. (1998). WMO solid
       precipitation measurement intercomparison (World Meteorological
       Organization, p. 212). World Meteorological Organization.
       http://www.wmo.int/pages/prog/www/reports/WMOtd872.pdf

    .. [2] Kochendorfer, J., Rasmussen, R., Wolff, M., Baker, B., Hall, M. E.,
       Meyers, T., Landolt, S., Jachcik, A., Isaksen, K., Brækkan, R., & Leeper, R.
       (2017). The quantification and correction of wind-induced precipitation
       measurement errors. Hydrology and Earth System Sciences, 21(4), 1973–1989.
       https://doi.org/10.5194/hess-21-1973-2017
    """
    precip_corr_configs = model.config.meteo.precipitation_correction

    if len(precip_corr_configs) == 0:
        return

    m = model.meteo
    dates = m.indexes['time']
    xs = m['x'].values
    ys = m['y'].values
    zs = m['alt'].values
    temps = m['temp'].values.T.copy()
    rel_hums = m['rel_hum'].values.T.copy()
    wind_speeds = m['wind_speed'].values.T.copy()
    precips = m['precip'].values.T  # no copy, because this will be modified in place

    # Interpolate missing temperature, humidity and wind speed values
    # TODO this is only really necessary for nonzero precipitation values; could be optimized
    _interpolate_temp_hum_wind(
        dates,
        xs,
        ys,
        zs,
        temps,
        rel_hums,
        wind_speeds,
        model.config.meteo.interpolation.temperature,
        model.config.meteo.interpolation.humidity,
        model.config.meteo.interpolation.wind,
    )

    temps_c = temps - constants.T0

    # Calculate precipitation phase (if required, calculate wet-bulb temperature first)
    precip_phase_method = model.config.meteo.precipitation_phase.method
    if precip_phase_method == 'temp':
        pp_temp = temps
    elif precip_phase_method == 'wet_bulb_temp':
        atmos_presss = meteo.atmospheric_pressure(zs)
        vap_presss = meteo.vapor_pressure(temps, rel_hums)
        spec_hums = meteo.specific_humidity(atmos_presss, vap_presss)
        spec_heat_caps_moist_air = meteo.specific_heat_capacity_moist_air(spec_hums)
        lat_heat_vaps = meteo.latent_heat_of_vaporization(temps)
        psych_consts = meteo.psychrometric_constant(
            atmos_presss,
            spec_heat_caps_moist_air,
            lat_heat_vaps,
        )
        wet_bulb_temps = meteo.wet_bulb_temperature(
            temps,
            rel_hums,
            vap_presss,
            psych_consts,
        )
        pp_temp = wet_bulb_temps

    snowfall_frac = meteo.precipitation_phase(
        pp_temp,
        threshold_temp=model.config.meteo.precipitation_phase.threshold_temp,
        temp_range=model.config.meteo.precipitation_phase.temp_range,
    )

    pos_precip = precips > 0
    pos_snow = pos_precip & (snowfall_frac == 1)
    pos_mixed = pos_precip & (snowfall_frac > 0) & (snowfall_frac < 1)

    for config in precip_corr_configs:
        method = config['method']

        if method not in (
            'constant_scf',
            'wmo',
            'kochendorfer',
        ):
            continue

        logger.info(f'Correcting station precipitation with method: {method}')

        if method == 'constant_scf':
            precips *= snowfall_frac * config['scf']
        elif method == 'wmo':
            gauge = config['gauge']
            cr = np.full(precips.shape, 100.)

            # Calculate wind speed at gauge height
            wind_speeds_2m = meteo.log_wind_profile(
                wind_speeds,
                model.config.meteo.measurement_height.wind,
                model.config.meteo.measurement_height.temperature,
                model.config.snow.roughness_length,
            )

            # Correction functions have been derived for wind speeds < 7 m s-1, so we do not allow
            # higher values here
            wind_speeds_2m = wind_speeds_2m.clip(max=7)

            # Some correction functions from [1] use Tmin and Tmax values; these have been replaced
            # by standard air temperature in the following.
            if gauge == 'nipher':
                cr[pos_snow] = (
                    100
                    - 0.44 * wind_speeds_2m[pos_snow]**2
                    - 1.98 * wind_speeds_2m[pos_snow]
                )
                cr[pos_mixed] = (
                    97.29
                    - 3.18 * wind_speeds_2m[pos_mixed]
                    + (0.58 - 0.67) * temps_c[pos_mixed]
                )
            elif gauge == 'tretyakov':
                cr[pos_snow] = 103.11 - 8.67 * wind_speeds_2m[pos_snow] + 0.30 * temps_c[pos_snow]
                cr[pos_mixed] = (
                    96.99
                    - 4.46 * wind_speeds_2m[pos_mixed]
                    + (0.88 + 0.22) * temps_c[pos_mixed]
                )
            elif gauge == 'us_sh':
                cr[pos_snow] = np.exp(4.61 - 0.04 * wind_speeds_2m[pos_snow]**1.75)
                cr[pos_mixed] = 101.04 - 5.62 * wind_speeds_2m[pos_mixed]
            elif gauge == 'us_unsh':
                cr[pos_snow] = np.exp(4.61 - 0.16 * wind_speeds_2m[pos_snow]**1.28)
                cr[pos_mixed] = 100.77 - 8.34 * wind_speeds_2m[pos_mixed]
            if gauge == 'hellmann':
                cr[pos_snow] = (
                    100
                    + 1.13 * wind_speeds_2m[pos_snow]**2
                    - 19.45 * wind_speeds_2m[pos_snow]
                )
                cr[pos_mixed] = (
                    96.63
                    + 0.41 * wind_speeds_2m[pos_mixed]**2
                    - 9.84 * wind_speeds_2m[pos_mixed]
                    + 5.95 * temps_c[pos_mixed]
                )
            else:
                raise NotImplementedError(f'Unknown gauge: {gauge}')
            
            precips /= (cr / 100.)
        elif method == 'kochendorfer':
            gauge = config['gauge']
            cr = np.full(precips.shape, 1.)  # CR is here not in percent but a fraction

            try:
                coeff_a, coeff_b, coeff_c, max_wind_speed = KOCHENDORFER_COEFFS[gauge]
            except KeyError:
                raise NotImplementedError(f'Unknown gauge: {gauge}')

            # Calculate catch ratio following [2] (use eq. (1) from the corrigendum, not eq. (4)
            # from the original paper)
            cr[pos_precip] = np.exp(
                -coeff_a * wind_speeds[pos_precip].clip(max=max_wind_speed)
                * (1 - np.arctan(coeff_b * temps_c[pos_precip]) + coeff_c)
            )

            precips /= cr


def _interpolate_temp_hum_wind(
    dates,
    xs,
    ys,
    zs,
    temps,
    rel_hums,
    wind_speeds,
    temp_config,
    hum_config,
    wind_speed_config,
):
    """
    Interpolate air temperature, relative humidity and wind speed recordings
    from the meteorological stations to the locations of the stations missing
    these recordings.
    """
    num_stations = len(xs)

    for date_num, date in enumerate(dates):
        nan_pos = np.isnan(temps[date_num, :])
        num_nan = nan_pos.sum()
        if num_nan > 0 and num_nan < num_stations:
            temps[date_num, nan_pos] = meteo.interpolate_param(
                'temp',
                date,
                temp_config,
                temps[date_num, ~nan_pos],
                xs[~nan_pos],
                ys[~nan_pos],
                zs[~nan_pos],
                xs[nan_pos],
                ys[nan_pos],
                zs[nan_pos],
            )

        nan_pos = np.isnan(rel_hums[date_num, :])
        num_nan = nan_pos.sum()
        if num_nan > 0 and num_nan < num_stations:
            rel_hums[date_num, nan_pos] = meteo.interpolate_param(
                'rel_hum',
                date,
                hum_config,
                rel_hums[date_num, ~nan_pos],
                xs[~nan_pos],
                ys[~nan_pos],
                zs[~nan_pos],
                xs[nan_pos],
                ys[nan_pos],
                zs[nan_pos],
                temps=temps[date_num, ~nan_pos],
                target_temps=temps[date_num, nan_pos],
            )

        nan_pos = np.isnan(wind_speeds[date_num, :])
        num_nan = nan_pos.sum()
        if num_nan > 0 and num_nan < num_stations:
            wind_speeds[date_num, nan_pos] = meteo.interpolate_param(
                'wind_speed',
                date,
                temp_config,
                wind_speeds[date_num, ~nan_pos],
                xs[~nan_pos],
                ys[~nan_pos],
                zs[~nan_pos],
                xs[nan_pos],
                ys[nan_pos],
                zs[nan_pos],
            )



from .atmosphere import (
    absolute_humidity,
    atmospheric_pressure,
    clear_sky_emissivity,
    cloud_factor_from_cloud_fraction,
    cloud_fraction_from_cloud_factor,
    cloud_fraction_from_humidity,
    dew_point_temperature,
    dry_air_density,
    latent_heat_of_vaporization,
    log_wind_profile,
    precipitable_water,
    precipitation_phase,
    pressure_to_altitude,
    psychrometric_constant,
    relative_humidity,
    saturation_vapor_pressure,
    specific_heat_capacity_moist_air,
    specific_humidity,
    vapor_pressure,
    wet_bulb_temperature,
    wind_from_uv,
    wind_to_uv,
)
from .interpolation import interpolate_station_data, interpolate_param
from .precipcorr import correct_station_precipitation



from .irradiance import (
    clear_sky_shortwave_irradiance,
    shortwave_irradiance,
    longwave_irradiance,
)
from .shadows import shadows
from .sunparams import (
    day_angle,
    declination_angle,
    equation_of_time,
    hour_angle,
    sun_vector,
    sun_parameters,
)



import numpy as np
from openamundsen import constants
from openamundsen.snowmodel import SnowModel
from .snow import (
    _compaction_anderson,
    fresh_snow_density,
    albedo,
    max_liquid_water_content,
    snow_properties,
)


class CryoLayerID:
    SNOW_FREE = -1
    NEW_SNOW = 0
    OLD_SNOW = 1
    FIRN = 2
    ICE = 3


class CryoLayerSnowModel(SnowModel):
    def __init__(self, model):
        s = model.state.snow
        num_cryo_layers = 4

        s.add_variable('num_layers', '1', 'Number of snow layers', dtype=int, retain=True)
        s.add_variable('thickness', 'm', 'Snow thickness', dim3=num_cryo_layers, retain=True)
        s.add_variable('density', 'kg m-3', 'Snow density', 'snow_density', dim3=num_cryo_layers, retain=True)
        s.add_variable('ice_content', 'kg m-2', 'Ice content of snow', dim3=num_cryo_layers, retain=True)
        s.add_variable('liquid_water_content', 'kg m-2', 'Liquid water content of snow', 'liquid_water_content_of_snow_layer', dim3=num_cryo_layers, retain=True)
        s.add_variable('cold_content', 'kg m-2', 'Cold content of snow', dim3=num_cryo_layers, retain=True)
        s.add_variable('temp', 'K', 'Snow temperature', dim3=num_cryo_layers, retain=True)  # TODO remove this (only added because this is in the default point outputs)
        s.add_variable('layer_albedo', '1', 'Snow layer albedo', dim3=num_cryo_layers, retain=True)
        s.add_variable('heat_cap', 'J K-1 m-2', 'Areal heat capacity of snow', dim3=num_cryo_layers)  # TODO remove this (only added because of snow.snow_properties())

        self.model = model
        self.num_cryo_layers = num_cryo_layers

    def initialize(self):
        roi = self.model.grid.roi
        s = self.model.state.snow

        s.swe[roi] = 0
        s.depth[roi] = 0
        s.area_fraction[roi] = 0
        s.num_layers[roi] = 0
        s.sublimation[roi] = 0
        s.thickness[:, roi] = 0
        s.ice_content[:, roi] = 0
        s.liquid_water_content[:, roi] = 0
        s.cold_content[:, roi] = 0
        s.density[:, roi] = np.nan
        s.layer_albedo[:, roi] = np.nan
        # s.temp[:, roi] = np.nan

    def albedo_aging(self):
        model = self.model
        s = model.state

        self.update_surface_layer_type()

        pos_old = s.snow.thickness[CryoLayerID.OLD_SNOW, :] > 0.
        pos_new = s.snow.thickness[CryoLayerID.NEW_SNOW, :] > 0.
        s.snow.albedo[pos_old] = s.snow.layer_albedo[CryoLayerID.OLD_SNOW, pos_old]
        s.snow.albedo[pos_new] = s.snow.layer_albedo[CryoLayerID.NEW_SNOW, pos_new]
        albedo(model, pos_old | pos_new)

        if model.config.snow.cryolayers.use_single_snow_albedo:
            s.snow.layer_albedo[CryoLayerID.OLD_SNOW, pos_old] = s.snow.albedo[pos_old]
            s.snow.layer_albedo[CryoLayerID.NEW_SNOW, pos_new] = s.snow.albedo[pos_new]
        else:
            pos_old_surf = s.surface.layer_type == CryoLayerID.OLD_SNOW
            pos_new_surf = s.surface.layer_type == CryoLayerID.NEW_SNOW
            s.snow.layer_albedo[CryoLayerID.OLD_SNOW, pos_old_surf] = s.snow.albedo[pos_old_surf]
            s.snow.layer_albedo[CryoLayerID.NEW_SNOW, pos_new_surf] = s.snow.albedo[pos_new_surf]

        # Firn and ice albedo stay constant
        if self.num_cryo_layers > 2:
            for i in (CryoLayerID.FIRN, CryoLayerID.ICE):
                pos = s.surface.layer_type == i

                if i == CryoLayerID.FIRN:
                    s.snow.layer_albedo[i, pos] = model.config.snow.albedo.firn
                elif i == CryoLayerID.ICE:
                    s.snow.layer_albedo[i, pos] = model.config.snow.albedo.ice

    def compaction(self):
        model = self.model
        s = model.state

        # Consider only old snow and new snow, but always set the number of layers to 2 regardless
        # if both layers exist, due to the loop over the layers in _compaction_anderson (and the
        # fact that an old snow layer could exist without a new snow layer (so using 1 as the number
        # of layers would not work here)
        num_layers = np.full(model.grid.roi.shape, 2)

        if model.config.snow.compaction.method == 'anderson':
            _compaction_anderson(
                model.grid.roi_idxs,
                model.timestep,
                num_layers,
                s.snow.thickness,
                s.snow.ice_content,
                s.snow.liquid_water_content,
                s.snow.density,
                model.state.meteo.temp,
            )
        else:
            raise NotImplementedError

        # Now handle firn and ice
        if self.num_cryo_layers > 2:
            ice_density = model.config.snow.cryolayers.transition.ice

            # Firn: linear transition to ice in ~10 yrs
            firnies = s.snow.thickness[CryoLayerID.FIRN, :] > 0.
            densification_rate = (
                (ice_density - 500.)  # old snow->firn transition density is ~500 kg m-3
                / (constants.HOURS_PER_DAY * constants.DAYS_PER_YEAR * 10)  # 10 years
            )
            s.snow.density[CryoLayerID.FIRN, firnies] += densification_rate * (model.timestep / constants.SECONDS_PER_HOUR)

            # Ice: density stays constant
            icies = s.snow.thickness[CryoLayerID.ICE, :] > 0.
            s.snow.density[CryoLayerID.ICE, icies] = ice_density

            total_we = s.snow.ice_content + s.snow.liquid_water_content
            s.snow.thickness[CryoLayerID.FIRN, firnies] = (
                total_we[CryoLayerID.FIRN, firnies] / s.snow.density[CryoLayerID.FIRN, firnies]
            )
            s.snow.thickness[CryoLayerID.ICE, icies] = (
                total_we[CryoLayerID.ICE, icies] / s.snow.density[CryoLayerID.ICE, icies]
            )

    def accumulation(self):
        model = self.model
        s = model.state
        pos = s.meteo.snowfall > 0
        self.add_snow(
            pos,
            s.meteo.snowfall[pos],
            density=fresh_snow_density(s.meteo.wet_bulb_temp[pos]),
        )

    def heat_conduction(self):
        pass

    def temperature_index(self):
        """
        Calculate snowmelt based on a temperature index approach.
        """
        model = self.model
        melt_config = model.config.snow.melt
        s = model.state

        timestep_d = model.timestep / (constants.SECONDS_PER_HOUR * constants.HOURS_PER_DAY)  # (d)

        snowies = s.snow.thickness.sum(axis=0) > 0.
        degree_days = (s.meteo.temp[snowies] - melt_config.threshold_temp) * timestep_d

        if melt_config.method == 'temperature_index':
            melt = melt_config.degree_day_factor * degree_days.clip(min=0)
        elif melt_config.method == 'enhanced_temperature_index':
            melt = (
                melt_config.degree_day_factor * degree_days
                + melt_config.albedo_factor
                * (1. - s.surface.albedo[snowies])
                * s.meteo.sw_in[snowies]
                * timestep_d
            )
            melt[degree_days <= 0.] = 0.
        else:
            raise NotImplementedError

        s.snow.melt[model.grid.roi] = 0.
        s.snow.melt[snowies] = np.minimum(
            melt.clip(min=0),
            s.snow.ice_content[:, snowies].sum(axis=0),
        )

    def melt(self):
        model = self.model
        s = model.state

        ice_content_change = s.snow.melt.copy()

        for i in range(self.num_cryo_layers):
            pos = (ice_content_change > 0) & (s.snow.ice_content[i, :] > 0)

            cur_ice_content_change = np.minimum(
                ice_content_change[pos],
                s.snow.ice_content[i, pos],
            )
            s.snow.thickness[i, pos] *= (1 - cur_ice_content_change / s.snow.ice_content[i, pos])
            s.snow.ice_content[i, pos] -= cur_ice_content_change
            s.snow.liquid_water_content[i, pos] += cur_ice_content_change
            ice_content_change[pos] -= cur_ice_content_change

    def sublimation(self):
        model = self.model
        roi = model.grid.roi
        s = model.state
        snowies_roi = s.snow.ice_content[:, roi].sum(axis=0) > 0.
        snowies = model.roi_mask_to_global(snowies_roi)

        s.snow.sublimation[roi] = 0.
        pot_sublim = -1 * s.surface.moisture_flux[snowies] * model.timestep
        pot_sublim[np.isnan(pot_sublim)] = 0.
        s.snow.sublimation[snowies] = np.minimum(pot_sublim, s.snow.ice_content[:, snowies].sum(axis=0))

        # Ice content change is positive for sublimation (mass loss) and negative for mass gain
        # (resublimation)
        ice_content_change = s.snow.sublimation.copy()

        # Sublimation: remove snow from the top layer
        for i in range(self.num_cryo_layers):
            pos = (ice_content_change > 0) & (s.snow.ice_content[i, :] > 0)

            cur_ice_content_change = np.minimum(
                ice_content_change[pos],
                s.snow.ice_content[i, pos],
            )
            s.snow.thickness[i, pos] *= (1 - cur_ice_content_change / s.snow.ice_content[i, pos])
            s.snow.ice_content[i, pos] -= cur_ice_content_change
            ice_content_change[pos] -= cur_ice_content_change

        # Resublimation: add snow to the top layer
        pos = (ice_content_change < 0)
        self.add_snow(
            pos,
            -ice_content_change[pos],
            density=fresh_snow_density(s.meteo.wet_bulb_temp[pos]),
        )

    def runoff(self):
        model = self.model
        s = model.state

        max_lwc = max_liquid_water_content(model)
        max_lwc[2:, :] = 0.  # no LWC for firn and ice

        runoff = model.state.meteo.rainfall.copy()
        runoff[np.isnan(runoff)] = 0.

        for i in range(self.num_cryo_layers):
            pos = (s.snow.ice_content[i, :] + s.snow.liquid_water_content[i, :]) > 0.

            s.snow.liquid_water_content[i, pos] += runoff[pos]
            runoff_cur = (s.snow.liquid_water_content[i, pos] - max_lwc[i, pos]).clip(min=0)
            runoff[pos] = runoff_cur
            s.snow.liquid_water_content[i, pos] -= runoff_cur

        s.snow.runoff[:] = runoff

    def update_layers(self):
        model = self.model
        s = model.state.snow
        transition_params = model.config.snow.cryolayers.transition

        total_we = s.ice_content + s.liquid_water_content

        # Reset empty layers
        for i in range(self.num_cryo_layers):
            self.reset_layer(i, total_we[i, :] <= 0.)

        # Update thickness
        pos = total_we > 0.
        s.thickness[pos] = total_we[pos] / s.density[pos]

        # Transition new snow -> old snow
        self.layer_transition(
            CryoLayerID.NEW_SNOW,
            CryoLayerID.OLD_SNOW,
            s.density[CryoLayerID.NEW_SNOW, :] >= transition_params.old_snow
        )

        # Transition old snow -> firn at the first timestep of the "transition month"
        if (
            model.date.month == transition_params.firn
            and model.date_idx > 0
            and model.dates[model.date_idx - 1].month != transition_params.firn
        ):
            self.layer_transition(CryoLayerID.OLD_SNOW, CryoLayerID.FIRN)

        # Transition firn -> ice
        self.layer_transition(
            CryoLayerID.FIRN,
            CryoLayerID.ICE,
            s.density[CryoLayerID.FIRN, :] >= transition_params.ice,
        )

        # Update snow depth (might have changed due to density changes) and number of layers
        s.depth[:] = s.thickness.sum(axis=0)
        s.num_layers[:] = (s.thickness > 0).sum(axis=0)

    def update_properties(self):
        model = self.model
        s = model.state

        snow_properties(self.model)

        s.snow.albedo[:] = np.nan

        for i in reversed(range(self.num_cryo_layers)):
            pos = s.snow.thickness[i, :] > 0
            s.snow.albedo[pos] = s.snow.layer_albedo[i, pos]
            s.snow.density[i, pos] = (
                s.snow.ice_content[i, pos] + s.snow.liquid_water_content[i, pos]
            ) / s.snow.thickness[i, pos]

    def reset_layer(self, layer, pos=None):
        s = self.model.state.snow

        if pos is None:
            pos = slice(None)

        s.thickness[layer, pos] = 0
        s.ice_content[layer, pos] = 0
        s.liquid_water_content[layer, pos] = 0
        s.cold_content[layer, pos] = 0
        s.density[layer, pos] = np.nan
        s.layer_albedo[layer, pos] = np.nan

    def layer_transition(self, src_layer, dst_layer, pos=None):
        s = self.model.state.snow

        if pos is None:
            pos = self.model.grid.roi

        pos_src_snow = pos & (s.thickness[src_layer, :] > 0.)
        pos_merge = pos_src_snow & (s.thickness[dst_layer, :] > 0.)
        pos_init = pos_src_snow & (s.thickness[dst_layer, :] == 0.)

        # Initialize the destination layer for pixels where SWE = 0
        s.thickness[dst_layer, pos_init] = s.thickness[src_layer, pos_init]
        s.ice_content[dst_layer, pos_init] = s.ice_content[src_layer, pos_init]
        s.liquid_water_content[dst_layer, pos_init] = s.liquid_water_content[src_layer, pos_init]
        s.cold_content[dst_layer, pos_init] = s.cold_content[src_layer, pos_init]
        s.density[dst_layer, pos_init] = s.density[src_layer, pos_init]
        s.layer_albedo[dst_layer, pos_init] = s.layer_albedo[src_layer, pos_init]

        # Take weighted mean for pixels where destination layer already exists
        w1 = s.ice_content[src_layer, pos_merge] + s.liquid_water_content[src_layer, pos_merge]
        w2 = s.ice_content[dst_layer, pos_merge] + s.liquid_water_content[dst_layer, pos_merge]

        # Normalize weights
        sum_weights = w1 + w2
        w1 /= sum_weights
        w2 /= sum_weights

        s.ice_content[dst_layer, pos_merge] += s.ice_content[src_layer, pos_merge]
        s.liquid_water_content[dst_layer, pos_merge] += s.liquid_water_content[src_layer, pos_merge]
        s.cold_content[dst_layer, pos_merge] += s.cold_content[src_layer, pos_merge]
        s.density[dst_layer, pos_merge] = (
            w1 * s.density[src_layer, pos_merge]
            + w2 * s.density[dst_layer, pos_merge]
        )
        s.layer_albedo[dst_layer, pos_merge] = (
            w1 * s.layer_albedo[src_layer, pos_merge]
            + w2 * s.layer_albedo[dst_layer, pos_merge]
        )
        s.thickness[dst_layer, pos_merge] = (
            (s.ice_content[dst_layer, pos_merge] + s.liquid_water_content[dst_layer, pos_merge])
            / s.density[dst_layer, pos_merge]
        )

        self.reset_layer(src_layer, pos)

    def update_surface_layer_type(self):
        model = self.model
        roi = model.grid.roi
        s = model.state

        s.surface.layer_type[roi] = CryoLayerID.SNOW_FREE

        for i in reversed(range(self.num_cryo_layers)):
            s.surface.layer_type[s.snow.thickness[i, :] > 0] = i

    def add_snow(
            self,
            pos,
            ice_content,
            liquid_water_content=0,
            density=None,
            albedo=None,
    ):
        model = self.model
        s = model.state

        pos_init = (s.snow.ice_content[CryoLayerID.NEW_SNOW, pos] == 0) & (ice_content > 0)
        pos_init_global = model.global_mask(pos_init, pos)

        # Initialize new snow layer where required
        if np.isscalar(density):
            density = np.full(ice_content.shape, density)
        s.snow.density[CryoLayerID.NEW_SNOW, pos_init_global] = density[pos_init]

        if albedo is None:
            s.snow.layer_albedo[CryoLayerID.NEW_SNOW, pos_init_global] = model.config.snow.albedo.max

            if model.config.snow.cryolayers.use_single_snow_albedo:
                pos_albedo = pos_init_global & (s.snow.thickness[CryoLayerID.OLD_SNOW, :] > 0.)
                s.snow.layer_albedo[CryoLayerID.NEW_SNOW, pos_albedo] = (
                    s.snow.layer_albedo[CryoLayerID.OLD_SNOW, pos_albedo]
                )

            s.snow.albedo[pos_init_global] = s.snow.layer_albedo[CryoLayerID.NEW_SNOW, pos_init_global]
        else:
            s.snow.layer_albedo[CryoLayerID.NEW_SNOW, pos] = albedo
            s.snow.albedo[pos] = albedo

        # Add snow to new snow layer
        s.snow.ice_content[CryoLayerID.NEW_SNOW, pos] += ice_content
        s.snow.liquid_water_content[CryoLayerID.NEW_SNOW, pos] += liquid_water_content
        s.snow.thickness[CryoLayerID.NEW_SNOW, pos] += ice_content / density
        s.snow.density[CryoLayerID.NEW_SNOW, pos] = (
            (
                s.snow.ice_content[CryoLayerID.NEW_SNOW, pos]
                + s.snow.liquid_water_content[CryoLayerID.NEW_SNOW, pos]
            ) / s.snow.thickness[CryoLayerID.NEW_SNOW, pos]
        )



import numpy as np
from numpy.testing import assert_allclose
import openamundsen.modules.radiation as rad
import pandas as pd
from pvlib import solarposition


def test_equation_of_time():
    doys = np.arange(366) + 1
    eot_oa = rad.equation_of_time(doys)
    eot_pv = solarposition.equation_of_time_spencer71(doys)
    assert_allclose(eot_pv, eot_oa, atol=0.15)  # error is due to 365 vs 365.25 used in calculating day angle


def test_declination_angle():
    doys = np.arange(366) + 1
    dec_oa = rad.declination_angle(doys)
    dec_pv = np.rad2deg(solarposition.declination_spencer71(doys))
    assert_allclose(dec_pv, dec_oa, atol=0.35)


def test_hour_angle():
    lon = 11.400375
    timezone = 1

    dates = pd.DatetimeIndex([
        '2020-01-02 08:43:37',
        '2020-07-15 16:55:31',
        '2020-09-01 12:23:11',
        '2020-11-15 23:40:00',
        '2020-12-24 15:34:11',
    ]).tz_localize(f'Etc/GMT-{timezone}')
    # (Etc/GMT-1 corresponds to GMT+1 in the traditional sense
    # (https://stackoverflow.com/questions/4008960/pytz-and-etc-gmt-5))

    eots = rad.equation_of_time(np.array([date.dayofyear for date in dates]))
    has_oa = rad.hour_angle(dates, timezone, lon, eots)
    has_pv = solarposition.hour_angle(dates, lon, eots)
    assert_allclose(has_pv, has_oa, atol=0.01)



import openamundsen as oa
import numpy as np


try:
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.use('QtAgg')
    mpl_available = True
except ImportError:
    mpl_available = False


class LiveView:
    """
    Class for managing a live view window of a model run.

    Parameters
    ----------
    config : dict
        The "liveview" section of the model run configuration (i.e.,
        model.config.liveview).

    state : StateVariableManager
        StateVariableManager (i.e., the model.state attribute) of the
        respective OpenAmundsen instance.

    roi : ndarray
        Boolean array specifying the region of interest.
    """
    def __init__(self, config, state, roi):
        if not mpl_available:
            raise ModuleNotFoundError(
                'matplotlib and a Qt binding (PyQt6/PySide6/PyQt5/PySide2) must be installed for '
                'the live view functionality'
            )

        self.config = config
        self.state = state
        self.roi = roi

    def create_window(self):
        """
        Prepare and open the live view window.
        """
        plt.style.use('dark_background')
        plt.rcParams['toolbar'] = 'None'

        num_cols = self.config.cols
        num_rows = int(np.ceil(len(self.config.variables) / num_cols))
        fig, axarr = plt.subplots(ncols=num_cols, nrows=num_rows)
        fig.subplots_adjust(top=0.92, bottom=0.03, left=0.03, right=0.97)

        fields = []
        imgs = []

        for field_num, d in enumerate(self.config.variables):
            field = d['var']
            min_range = d['min']
            max_range = d['max']

            data = self.state[field]

            ax = axarr.flat[field_num]
            img = ax.imshow(
                data,
                vmin=min_range,
                vmax=max_range,
                interpolation='None',
                cmap=self.config.cmap,
            )

            cbar = fig.colorbar(img, ax=ax)
            cbar.ax.tick_params(labelsize='x-small')

            ax.set_title(self._var_label(field), fontsize='small')
            ax.axis('off')

            imgs.append(img)
            fields.append(field)

        # Hide unused subplot spaces
        for ax in axarr.flat[len(self.config.variables):]:
            ax.axis('off')

        plt.text(
            x=0.03,
            y=0.95,
            s=f'openAMUNDSEN v{oa.__version__}',
            fontsize='xx-large',
            ha='left',
            transform=fig.transFigure,
        )
        time_label = plt.text(
            x=0.97,
            y=0.95,
            s='',
            fontsize='large',
            ha='right',
            transform=fig.transFigure,
            bbox=dict(  # set the text background color to ensure that old values are overplotted
                facecolor=plt.rcParams['figure.facecolor'],
                linewidth=0,
            ),
        )

        self.fields = fields
        self.fig = fig
        self.imgs = imgs
        self.axarr = axarr
        self.time_label = time_label

        if self.config.blit:
            fig.canvas.mpl_connect('draw_event', self._on_draw)

        plt.show(block=False)
        fig.canvas.draw()
        mgr = plt.get_current_fig_manager()
        mgr.set_window_title('openAMUNDSEN')
        mgr.resize(self.config.width, self.config.height)

    def update(self, date):
        """
        Update the live view window with the current values of the state
        variables.

        Parameters
        ----------
        date : datetime
            Current model time step.
        """
        fig = self.fig

        if self.config.blit:
            fig.canvas.restore_region(self.fig_bg)

        # Update time label
        time_label = self.time_label
        time_label.set_text(f'{date:%Y-%m-%d %H:%M}')

        if self.config.blit:
            fig.draw_artist(time_label)
            label_extent = time_label.get_window_extent(renderer=fig.canvas.get_renderer())
            label_bbox = matplotlib.transforms.TransformedBbox(
                label_extent.transformed(fig.transFigure.inverted()),
                fig.transFigure,
            )
            fig.canvas.blit(label_bbox)

        # Update images
        for field, ax, img in zip(self.fields, self.axarr.flat, self.imgs):
            data = self.state[field]

            # Hide non-ROI pixels (conversion to float is required for integer fields)
            data = data.astype(float, copy=False)
            data[~self.roi] = np.nan

            # Perform downsampling
            data = data[::self.config.downsample]

            img.set_data(data)

            if self.config.blit:
                ax.draw_artist(img)
                fig.canvas.blit(ax.bbox)

        if not self.config.blit:
            fig.canvas.draw()

        fig.canvas.flush_events()

    def close(self):
        """
        Close the window.
        """
        plt.close(self.fig)

    def __del__(self):
        """
        Close the window when the object is deleted.
        """
        self.close()

    def _on_draw(self, event):
        fig = self.fig
        self.fig_bg = fig.canvas.copy_from_bbox(fig.bbox)

    def _var_label(self, field):
        """
        Return the label for a variable to be shown in the live view window.

        Parameters
        ----------
        field : str
            "<category>.<var_name>" string.

        Returns
        -------
        label : str
        """
        category, var_name = self.state.parse(field)
        meta = self.state[category]._meta[var_name]

        if meta.long_name is not None:
            label = meta.long_name
        elif meta.standard_name is not None:
            label = meta.standard_name
        else:
            label = var_name

        if meta.units is not None:
            label += f'\n({meta.units})'

        return label



import numpy as np
import openamundsen.constants as c
import pandas as pd


def day_angle(doy):
    """
    Return the day of the year in angular form.

    Parameters
    ----------
    doy : int
        Day of the year (Jan 1 = 1).

    Returns
    -------
    day_angle : float
        Day angle in radians.
    """
    # TODO
    #   Use 2*pi/365 for normal years and 2*pi/366 for leap years
    #   instead of 365.25?
    return (2. * np.pi / c.DAYS_PER_YEAR) * (doy - 1)


def equation_of_time(doy, method='spencer'):
    """
    Calculate the equation of time, i.e., the difference in time between solar
    noon at 0 degrees longitude and 12:00 UTC.

    Parameters
    ----------
    doy : int
        Day of the year (Jan 1 = 1).

    Returns
    -------
    eot_m : float
        Equation of time in minutes for the given day. The value is within
        +/- 16 minutes throughout the entire year.

    References
    ----------
    .. [1] J. W. Spencer, "Fourier series representation of the position of the
       sun" in Search 2 (5), p. 172 (1971)
    """
    # TODO
    #   Implement Reda method (https://www.nrel.gov/docs/fy08osti/34302.pdf),
    #   possibly PVCDROM (see https://github.com/pvlib/pvlib-python/blob/master/pvlib/solarposition.py)

    da = day_angle(doy)

    if method == 'spencer':
        eot = (
            0.0000075 +
            0.001868 * np.cos(da) - 0.032077 * np.sin(da) -
            0.014615 * np.cos(2 * da) - 0.040849 * np.sin(2 * da)
        )  # in radians
    else:
        raise NotImplementedError(f'Unsupported method: {method}')

    eot_m = c.HOURS_PER_DAY * c.MINUTES_PER_HOUR / (2 * np.pi) * eot  # in minutes
    return eot_m


def declination_angle(doy):
    """
    Calculate the solar declination angle after Bourges (1985).

    Parameters
    ----------
    doy : int
        Day of the year (Jan 1 = 1).

    Returns
    -------
    declination : float
        Solar declination angle in degrees.

    References
    ----------
    .. [1] Bernard Bourges. Improvement in solar declination computation. Solar
       Energy, 1985, 35(4), pp.367-369.
    """
    pass
    
    day_number = np.deg2rad((360 / c.DAYS_PER_YEAR) * (doy - 79.346))
    declination = (
        0.3723 + 23.2567 * np.sin(day_number) - 0.7580 * np.cos(day_number)
        + 0.1149 * np.sin(2 * day_number) + 0.3656 * np.cos(2 * day_number)
        - 0.1712 * np.sin(3 * day_number) + 0.0201 * np.cos(3 * day_number)
    )
    return declination


def hour_angle(date, timezone, lon, eot):
    """
    Calculate the hour angle, i.e., the angular displacement of the sun east or
    west of the local meridian due to rotation of the earth on its axis at
    15° per hour.

    Parameters
    ----------
    date : datetime-like
        Local date and time.

    timezone : int
        Timezone, e.g. 1 for CET.

    lon : float
        Longitude.

    eot : float
        Equation of time in minutes.

    Returns
    -------
    hour_angle : float
        Hour angle in degrees.
    """
    date = pd.to_datetime(date)
    hour = (date - date.normalize()).total_seconds() / c.SECONDS_PER_HOUR  # fractional hour of the day

    lstm = c.STANDARD_TIMEZONE_WIDTH * timezone  # local standard time meridian
    tc = c.MINUTES_PER_DEGREE_OF_EARTH_ROTATION * (lon - lstm) + eot  # time correction (minutes)
    lst = hour + tc / c.MINUTES_PER_HOUR  # local solar time
    ha = c.SUN_DEGREES_PER_HOUR * (lst - 12)

    return ha


def sun_vector(lat, ha, dec):
    """
    Calculate the vector defining the position of the sun after Corripio
    (2003).

    Parameters
    ----------
    lat : float
        Latitude (degrees).

    ha : float
        Hour angle (degrees).

    dec : float
        Declination angle (degrees).

    Returns
    -------
    vec : ndarray
        Solar vector.

    References
    ----------
    .. [1] Corripio, J. G. (2003). Vectorial algebra algorithms for calculating
       terrain parameters from DEMs and solar radiation modelling in mountainous
       terrain. International Journal of Geographical Information Science, 17(1),
       1–23. https://doi.org/10.1080/13658810210157796
    """
    lat_rad = np.deg2rad(lat)
    ha_rad = np.deg2rad(ha)
    dec_rad = np.deg2rad(dec)

    return np.array([
        -np.sin(ha_rad) * np.cos(dec_rad),
        np.sin(lat_rad) * np.cos(ha_rad) * np.cos(dec_rad) - np.cos(lat_rad) * np.sin(dec_rad),
        np.cos(lat_rad) * np.cos(ha_rad) * np.cos(dec_rad) + np.sin(lat_rad) * np.sin(dec_rad),
    ])


def sun_parameters(date, lon, lat, timezone):
    """
    Calculate sun related parameters for a specified date and position.

    Parameters
    ----------
    date : datetime-like
        Local date and time.

    lon : float
        Longitude (degrees).

    lat : float
        Latitude (degrees).

    timezone : int
        Timezone, e.g. 1 for CET.

    Returns
    -------
    d : dict
        Dictionary containing the following keys:
        - 'day_angle': day angle (radians)
        - 'hour_angle': hour angle (degrees)
        - 'declination_angle': declination angle (degrees)
        - 'equation_of_time': equation of time (minutes)
        - 'sun_vector': vector describing the position of the sun
        - 'zenith_angle': zenith angle (degrees)
        - 'sun_over_horizon': True if the sun is over the horizon
    """
    date = pd.to_datetime(date)
    eot = equation_of_time(date.dayofyear)
    da = day_angle(date.dayofyear)
    ha = hour_angle(date, timezone, lon, eot)
    dec = declination_angle(date.dayofyear)
    sv = sun_vector(lat, ha, dec)
    zenith_angle = np.rad2deg(np.arccos(sv[2]))

    return {
        'day_angle': da,
        'hour_angle': ha,
        'declination_angle': dec,
        'equation_of_time': eot,
        'sun_vector': sv,
        'zenith_angle': zenith_angle,
        'sun_over_horizon': zenith_angle < 90,
    }



from .conftest import base_config
import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
import openamundsen as oa
import openamundsen.errors as errors
import pandas as pd
import pytest
import xarray as xr


@pytest.mark.parametrize('fmt', ['netcdf', 'csv', 'memory'])
def test_formats(fmt, tmp_path):
    config = base_config()
    config.end_date = '2020-01-16'
    config.results_dir = tmp_path
    config.output_data.timeseries.format = fmt
    config.output_data.timeseries.variables = [{'var': 'snow.num_layers'}]

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()

    point_ids = ['bellavista', 'latschbloder', 'proviantdepot']

    if fmt in ('netcdf', 'memory'):
        if fmt == 'netcdf':
            # Set mask_and_scale=False because otherwise integer variables with a _FillValue are
            # converted to float by xarray (see e.g. https://github.com/pydata/xarray/issues/1194),
            # even when they do not actually contain any missing values.
            ds = xr.open_dataset(
                tmp_path / 'output_timeseries.nc',
                mask_and_scale=False,
            )
        elif fmt == 'memory':
            ds = model.point_output.data

        assert ds.time.to_index().equals(model.dates)
        assert_array_equal(ds.point, point_ids)
        assert_array_equal(
            list(ds.coords.keys()),
            ['time', 'point', 'lon', 'lat', 'alt', 'x', 'y', 'soil_layer', 'snow_layer'],
        )
        assert ds.temp.dims == ('time', 'point')
        assert ds.snow_thickness.dims == ('time', 'snow_layer', 'point')
        assert ds.soil_temp.dims == ('time', 'soil_layer', 'point')
        assert ds.temp.dtype == np.float32
        assert np.issubdtype(ds.num_layers.dtype, np.integer)
        assert np.all(ds.temp > 250.)
    elif fmt == 'csv':
        for point_id in point_ids:
            assert (tmp_path / f'point_{point_id}.csv').exists()

        df = pd.read_csv(tmp_path / 'point_bellavista.csv', index_col='time', parse_dates=True)
        assert df.index.equals(model.dates)
        assert df.temp.dtype == np.float64
        assert np.issubdtype(df.num_layers.dtype, np.integer)
        assert 'snow_thickness0' in df
        assert np.all(df.temp > 250.)


def test_values():
    config = base_config()
    config.end_date = '2020-01-15'

    model = oa.OpenAmundsen(config)
    model.initialize()

    point = 'proviantdepot'
    row = int(model.meteo.sel(station=point).row)
    col = int(model.meteo.sel(station=point).col)

    data_temp = pd.Series(index=model.dates, dtype=float)
    data_soil_temp1 = pd.Series(index=model.dates, dtype=float)

    for date in model.dates:
        model.run_single()
        data_temp[date] = model.state.meteo.temp[row, col]
        data_soil_temp1[date] = model.state.soil.temp[1, row, col]

    ds = model.point_output.data.sel(point=point)
    assert_allclose(ds.temp.values, data_temp)
    assert_allclose(ds.soil_temp.isel(soil_layer=1).values, data_soil_temp1)


@pytest.mark.parametrize('write_freq', ['ME', '7h', '3h', '10min'])
def test_write_freq(write_freq, tmp_path):
    config = base_config()
    config.end_date = '2020-01-15'
    config.results_dir = tmp_path
    config.output_data.timeseries.format = 'netcdf'
    config.output_data.timeseries.write_freq = write_freq

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()

    ds = xr.open_dataset(tmp_path / 'output_timeseries.nc')
    assert ds.time.to_index().equals(model.dates)


def test_points():
    bc = base_config()
    bc.end_date = '2020-01-15 00:00'

    config = bc.copy()
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert_array_equal(ds.point, ['bellavista', 'latschbloder', 'proviantdepot'])

    config = bc.copy()
    config.output_data.timeseries.add_default_points = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert ds.point.size == 0

    config = bc.copy()
    config.output_data.timeseries.add_default_points = False
    config.output_data.timeseries.points.append({
        'x': 640367,
        'y': 5182896,
    })
    config.output_data.timeseries.points.append({
        'x': 645378,
        'y': 5190907,
        'name': 'mypoint',
    })
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert_array_equal(ds.point, ['point1', 'mypoint'])
    assert_allclose(ds.alt, [3181.89, 1948.97])

    # Duplicate point name
    config = bc.copy()
    config.output_data.timeseries.points.append({
        'x': 640367,
        'y': 5182896,
        'name': 'bellavista',
    })
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.ConfigurationError):
        model.initialize()

    # Duplicate point name
    config = bc.copy()
    config.output_data.timeseries.points.append({
        'x': 640367,
        'y': 5182896,
        'name': 'bellavista',
    })
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.ConfigurationError):
        model.initialize()

    # Point not within grid
    config = bc.copy()
    config.output_data.timeseries.points.append({
        'x': 637152,
        'y': 5196427,
    })
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.ConfigurationError):
        model.initialize()


def test_variables():
    bc = base_config()
    bc.end_date = '2020-01-15 00:00'
    bc.output_data.timeseries.variables = []

    config = bc.copy()
    config.output_data.timeseries.add_default_variables = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert len(ds.data_vars) == 0

    config = bc.copy()
    config.output_data.timeseries.add_default_variables = False
    config.output_data.timeseries.variables.append({'var': 'meteo.spec_heat_cap_moist_air'})
    config.output_data.timeseries.variables.append({
        'var': 'surface.conductance',
        'name': 'myvar',
    })
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert_array_equal(ds.data_vars, ['spec_heat_cap_moist_air', 'myvar'])

    # Invalid variable name
    config = bc.copy()
    config.output_data.timeseries.variables.append({'var': 'meteo.asdf'})
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.ConfigurationError):
        model.initialize()

    # Output name already in use
    config = bc.copy()
    config.output_data.timeseries.variables.append({'var': 'meteo.temp'})
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.ConfigurationError):
        model.initialize()
